#include "NUHMSSMNoFVHimalaya_input_parameters.hpp"
#include "NUHMSSMNoFVHimalaya_mass_eigenstates.hpp"
#include "NUHMSSMNoFVHimalaya_observables.hpp"
#include "NUHMSSMNoFVHimalaya_slha_io.hpp"
#include "config.h"
#include "ew_input.hpp"
#include "lowe.h"

#include "NUHMSSMNoFVHimalaya_two_scale_initial_guesser.hpp"
#include "NUHMSSMNoFVHimalaya_two_scale_model.hpp"
#include "NUHMSSMNoFVHimalaya_two_scale_susy_scale_constraint.hpp"
#include "global_thread_pool.hpp"
#include "logger.hpp"
#include "loop_corrections.hpp"
#include "root_finder.hpp"
#include "standard_model.hpp"
#include "wrappers.hpp"
#include <cmath>

#include "mssm_twoloop_mt.hpp"
#include "sm_fourloophiggs.hpp"
#include "sm_threeloop_as.hpp"
#include "sm_threeloophiggs.hpp"
#include "sm_twoloophiggs.hpp"

#include "command_line_options.hpp"
#include "physical_input.hpp"
#include "rkf_integrator.hpp"
#include "spectrum_generator_settings.hpp"

#define HIGGS_2LOOP_CORRECTION_AT_AS get_loop_corrections().higgs_at_as
#define HIGGS_2LOOP_CORRECTION_AB_AS get_loop_corrections().higgs_ab_as
#define HIGGS_2LOOP_CORRECTION_AT_AT get_loop_corrections().higgs_at_at
#define HIGGS_2LOOP_CORRECTION_ATAU_ATAU get_loop_corrections().higgs_atau_atau
#define TOP_POLE_QCD_CORRECTION get_loop_corrections().top_qcd
#define HIGGS_3LOOP_CORRECTION_AT_AS_AS get_loop_corrections().higgs_at_as_as
#define HIGGS_3LOOP_CORRECTION_AB_AS_AS get_loop_corrections().higgs_ab_as_as
#define HIGGS_3LOOP_MDR_SCHEME get_loop_corrections().higgs_3L_mdr_scheme
#define HIGGS_3LOOP_CORRECTION_AT_AT_AS get_loop_corrections().higgs_at_at_as
#define HIGGS_3LOOP_CORRECTION_AT_AT_AT get_loop_corrections().higgs_at_at_at
#define HIGGS_4LOOP_CORRECTION_AT_AS_AS_AS                                     \
  get_loop_corrections().higgs_at_as_as_as

// 7 parameters to be varied: g_i + y_{t,b,tau} + v
static const int NPARS = 7;
static const double GAUGELESS =
    1e-10; // approximates the gaugeless limit for gauge couplings
using Vec_t = Eigen::Matrix<double, NPARS, 1>;

// root solver iteration counter
int Iteration = 0;
Vec_t high_scale_pars;
double abs_deviation;

namespace flexiblesusy {

/// fills relevant MSSM parameters from `vec'
void fill_parameters(NUHMSSMNoFVHimalaya_mass_eigenstates &mssm,
                     const Vec_t &vec) {
  int idx = 0;

  mssm.set_g1(vec(idx++));
  mssm.set_g2(vec(idx++));
  mssm.set_g3(vec(idx++));
  mssm.set_Yu(2, 2, vec(idx++));
  //  mssm.set_Yu(1,1,vec(idx++));
  mssm.set_Yd(2, 2, vec(idx++));
  //  mssm.set_Yd(1,1,vec(idx++));
  mssm.set_Ye(2, 2, vec(idx++));
  // mssm.set_Ye(1,1,vec(idx++));

  const double v = vec(idx++);
  const double tb = mssm.get_vu() == 0. || mssm.get_vd() == 0.
                        ? 1
                        : mssm.get_vu() / mssm.get_vd();

  mssm.set_vu(v * tb / Sqrt(1 + Sqr(tb)));
  mssm.set_vd(v * 1 / Sqrt(1 + Sqr(tb)));
}

void obtain_parameters(NUHMSSMNoFVHimalaya_mass_eigenstates &mssm, Vec_t &vec) {
  int idx = 0;

  vec(idx++) = mssm.get_g1();
  vec(idx++) = mssm.get_g2();
  vec(idx++) = mssm.get_g3();
  vec(idx++) = mssm.get_Yu(2, 2);
  vec(idx++) = mssm.get_Yd(2, 2);
  vec(idx++) = mssm.get_Ye(2, 2);
  vec(idx++) = Sqrt(Sqr(mssm.get_vu()) + Sqr(mssm.get_vd()));
}

/// creates vector of given MSSM parameters
Vec_t to_vec(double g1, double g2, double g3,
             const Eigen::Matrix<double, 3, 3> &Yu,
             const Eigen::Matrix<double, 3, 3> &Yd,
             const Eigen::Matrix<double, 3, 3> &Ye, double vev) {
  Vec_t vec;
  int idx = 0;

  vec(idx++) = g1;
  vec(idx++) = g2;
  vec(idx++) = g3;
  vec(idx++) = Yu(2, 2);
  vec(idx++) = Yd(2, 2);
  vec(idx++) = Ye(2, 2);
  vec(idx++) = vev;

  return vec;
}

/// creates vector of relevant SM parameters
Vec_t to_vec(const NUHMSSMNoFVHimalaya_mass_eigenstates &mssm) {
  return to_vec(mssm.get_g1(), mssm.get_g2(), mssm.get_g3(), mssm.get_Yu(),
                mssm.get_Yd(), mssm.get_Ye(),
                Sqrt(Sqr(mssm.get_vu()) + Sqr(mssm.get_vd())));
}

softsusy::QedQcd make_sm_input() {
  softsusy::QedQcd qedqcd;
  return qedqcd;
}

void fill_parameters(NUHMSSMNoFVHimalaya_mass_eigenstates &mssm,
                     const NUHMSSMNoFVHimalaya_input_parameters &input) {
  const auto Yu = mssm.get_Yu();
  const auto Yd = mssm.get_Yd();
  const auto Ye = mssm.get_Ye();
  const auto tb = input.TanBeta;
  const auto v = Sqrt(Sqr(mssm.get_vu()) + Sqr(mssm.get_vd()));

  Eigen::Matrix<double, 3, 3> TYu, TYd, TYe, mq2, mu2, md2, ml2, me2;
  TYu.setZero();
  TYd.setZero();
  TYe.setZero();
  mq2.setZero();
  mu2.setZero();
  md2.setZero();
  ml2.setZero();
  me2.setZero();

  TYu(0, 0) = input.AuIN * Yu(0, 0);
  TYu(1, 1) = input.AcIN * Yu(1, 1);
  TYu(2, 2) = input.AtIN * Yu(2, 2);
  TYd(0, 0) = input.AdIN * Yd(0, 0);
  TYd(1, 1) = input.AsIN * Yd(1, 1);
  TYd(2, 2) = input.AbIN * Yd(2, 2);
  TYe(0, 0) = input.AeIN * Ye(0, 0);
  TYe(1, 1) = input.AmuonIN * Ye(1, 1);
  TYe(2, 2) = input.AtauIN * Ye(2, 2);

  mq2 << Sqr(input.mq11IN), 0, 0, 0, Sqr(input.mq22IN), 0, 0, 0,
      Sqr(input.mq33IN);
  mu2 << Sqr(input.mu11IN), 0, 0, 0, Sqr(input.mu22IN), 0, 0, 0,
      Sqr(input.mu33IN);
  md2 << Sqr(input.md11IN), 0, 0, 0, Sqr(input.md22IN), 0, 0, 0,
      Sqr(input.md33IN);
  ml2 << Sqr(input.ml11IN), 0, 0, 0, Sqr(input.ml22IN), 0, 0, 0,
      Sqr(input.ml33IN);
  me2 << Sqr(input.me11IN), 0, 0, 0, Sqr(input.me22IN), 0, 0, 0,
      Sqr(input.me33IN);

  mssm.set_scale(input.Qin);
  mssm.set_vu(v * tb / Sqrt(1 + Sqr(tb)));
  mssm.set_vd(v * 1 / Sqrt(1 + Sqr(tb)));
  mssm.set_MassB(input.M1);
  mssm.set_MassWB(input.M2);
  mssm.set_MassG(input.M3);
  mssm.set_TYu(TYu);
  mssm.set_TYd(TYd);
  mssm.set_TYe(TYe);
  mssm.set_mq2(mq2);
  mssm.set_mu2(mu2);
  mssm.set_md2(md2);
  mssm.set_ml2(ml2);
  mssm.set_me2(me2);
  mssm.set_Mu(input.MuIN);
  mssm.set_BMu(input.mA2IN / (input.TanBeta + 1 / input.TanBeta));

  mssm.solve_ewsb_tree_level();
}

void fill_settings(NUHMSSMNoFVHimalaya_mass_eigenstates &model,
                   const Spectrum_generator_settings &settings) {
  model.set_pole_mass_loop_order(
      settings.get(Spectrum_generator_settings::pole_mass_loop_order));
  model.set_ewsb_loop_order(
      settings.get(Spectrum_generator_settings::ewsb_loop_order));
  model.set_loop_corrections(settings.get_loop_corrections());
  model.set_threshold_corrections(settings.get_threshold_corrections());

  model.do_calculate_sm_pole_masses(
      settings.get(Spectrum_generator_settings::calculate_sm_masses));
  model.do_calculate_bsm_pole_masses(
      settings.get(Spectrum_generator_settings::calculate_bsm_masses));
  model.do_force_output(
      settings.get(Spectrum_generator_settings::force_output));
  model.set_loops(settings.get(Spectrum_generator_settings::beta_loop_order));
  model.set_thresholds(settings.get(
      Spectrum_generator_settings::threshold_corrections_loop_order));
  model.set_zero_threshold(
      settings.get(Spectrum_generator_settings::beta_zero_threshold));
}

void fill_settings(standard_model::Standard_model &sm,
                   const Spectrum_generator_settings &settings) {
  sm.set_pole_mass_loop_order(
      settings.get(Spectrum_generator_settings::pole_mass_loop_order));
  sm.set_ewsb_loop_order(
      settings.get(Spectrum_generator_settings::ewsb_loop_order));
  sm.set_loop_corrections(settings.get_loop_corrections());
  sm.set_threshold_corrections(settings.get_threshold_corrections());

  sm.do_force_output(settings.get(Spectrum_generator_settings::force_output));
  sm.set_loops(settings.get(Spectrum_generator_settings::beta_loop_order));
  sm.set_thresholds(settings.get(
      Spectrum_generator_settings::threshold_corrections_loop_order));
  sm.set_zero_threshold(
      settings.get(Spectrum_generator_settings::beta_zero_threshold));
}

/// initial guess of SM parameters
NUHMSSMNoFVHimalaya_mass_eigenstates
initial_guess(const NUHMSSMNoFVHimalaya_input_parameters &mssm_input) {
  const auto tb = mssm_input.TanBeta;
  double MS = Sqrt(mssm_input.mq33IN * mssm_input.mu33IN);

  Eigen::Matrix<double, 3, 3> Yu, Yd, Ye;
  Yu.setZero();
  Yd.setZero();
  Ye.setZero();

  standard_model::Standard_model sm;
  sm.set_precision(1e-2);
  sm.set_loops(3);
  sm.set_ewsb_loop_order(2);

  sm.set_scale(Electroweak_constants::MZ);
  sm.set_g1(Electroweak_constants::g1);
  sm.set_g2(Electroweak_constants::g2);
  sm.set_g3(Electroweak_constants::g3);
  sm.set_v(Electroweak_constants::vSM);
  sm.set_Lambdax(Electroweak_constants::lamSM);

  sm.set_Yu(2, 2, Electroweak_constants::ytSM);
  sm.set_Yd(2, 2, Electroweak_constants::ybSM);
  sm.set_Ye(2, 2, Electroweak_constants::ylSM);

  sm.calculate_DRbar_masses();
  sm.set_integrator(runge_kutta::RKF_integrator());
  sm.run_to(MS);
  sm.calculate_DRbar_masses();

  NUHMSSMNoFVHimalaya_mass_eigenstates mssm;
  mssm.set_precision(1e-2);
  mssm.set_loops(3);
  mssm.set_pole_mass_loop_order(2);
  mssm.set_ewsb_loop_order(2);

  Eigen::Matrix<double, 3, 3> upQuarksDRbar = ZEROMATRIX(3, 3);
  Eigen::Matrix<double, 3, 3> downQuarksDRbar = ZEROMATRIX(3, 3);
  Eigen::Matrix<double, 3, 3> downLeptonsDRbar = ZEROMATRIX(3, 3);

  double delta_mt = 0.;

  double k = 1. / Sqr(4 * Pi);
  double gs2 = Sqr(sm.get_g3());
  double Xt = (mssm_input.AtIN - mssm_input.MuIN / tb) / MS;
  delta_mt +=
      sm.get_MFu(2) * k * gs2 * 4. / 3. * (Xt - 1.); /* gluino contribution
                                                      * only, good estimate
                                                      */
  upQuarksDRbar(2, 2) = sm.get_MFu(2) + delta_mt;
  downQuarksDRbar(2, 2) = sm.get_MFd(2);
  downLeptonsDRbar(2, 2) = sm.get_MFe(2);

  mssm.set_g1(sm.get_g1());
  mssm.set_g2(sm.get_g2());
  mssm.set_g3(sm.get_g3());
  mssm.set_vu(sm.get_v() * tb / Sqrt(1 + Sqr(tb)));
  mssm.set_vd(sm.get_v() * 1 / Sqrt(1 + Sqr(tb)));
  mssm.set_Yu((1.4142135623730951 * upQuarksDRbar) / mssm.get_vu());
  mssm.set_Yd((1.4142135623730951 * downQuarksDRbar) / mssm.get_vd());
  mssm.set_Ye((1.4142135623730951 * downLeptonsDRbar) / mssm.get_vd());

  fill_parameters(mssm, mssm_input);

  return mssm;
}

double
calculate_delta_alpha_em(double alpha_em,
                         const NUHMSSMNoFVHimalaya_mass_eigenstates &model) {
  const double currentScale = model.get_scale();
  double delta_alpha_em = 0.;

  const auto MCha = model.get_MCha();
  const auto MHpm = model.get_MHpm();
  const auto MSd = model.get_MSd();
  const auto MSe = model.get_MSe();
  const auto MSu = model.get_MSu();
  const auto MSs = model.get_MSs();
  const auto MSm = model.get_MSm();
  const auto MSc = model.get_MSc();
  const auto MSb = model.get_MSd();
  const auto MStau = model.get_MStau();
  const auto MSt = model.get_MSt();

  delta_alpha_em +=
      alpha_em / (2. * Pi) *
      (0.3333333333333333 -
       1.3333333333333333 * FiniteLog(Abs(MCha(0) / currentScale)) -
       1.3333333333333333 * FiniteLog(Abs(MCha(1) / currentScale)) -
       0.3333333333333333 * FiniteLog(Abs(MHpm(1) / currentScale)) -
       0.1111111111111111 * FiniteLog(Abs(MSd(0) / currentScale)) -
       0.1111111111111111 * FiniteLog(Abs(MSd(1) / currentScale)) -
       0.3333333333333333 * FiniteLog(Abs(MSe(0) / currentScale)) -
       0.3333333333333333 * FiniteLog(Abs(MSe(1) / currentScale)) -
       0.4444444444444444 * FiniteLog(Abs(MSu(0) / currentScale)) -
       0.4444444444444444 * FiniteLog(Abs(MSu(1) / currentScale)) -
       0.1111111111111111 * FiniteLog(Abs(MSs(0) / currentScale)) -
       0.1111111111111111 * FiniteLog(Abs(MSs(1) / currentScale)) -
       0.3333333333333333 * FiniteLog(Abs(MSm(0) / currentScale)) -
       0.3333333333333333 * FiniteLog(Abs(MSm(1) / currentScale)) -
       0.4444444444444444 * FiniteLog(Abs(MSc(0) / currentScale)) -
       0.4444444444444444 * FiniteLog(Abs(MSc(1) / currentScale)) -
       0.1111111111111111 * FiniteLog(Abs(MSb(0) / currentScale)) -
       0.1111111111111111 * FiniteLog(Abs(MSb(1) / currentScale)) -
       0.3333333333333333 * FiniteLog(Abs(MStau(0) / currentScale)) -
       0.3333333333333333 * FiniteLog(Abs(MStau(1) / currentScale)) -
       0.4444444444444444 * FiniteLog(Abs(MSt(0) / currentScale)) -
       0.4444444444444444 * FiniteLog(Abs(MSt(1) / currentScale)));

  return delta_alpha_em;
}

double
calculate_delta_alpha_s(double alpha_s,
                        const NUHMSSMNoFVHimalaya_mass_eigenstates &model) {
  const double currentScale = model.get_scale();
  double delta_alpha_s = 0.;

  const auto MSd = model.get_MSd();
  const auto MSu = model.get_MSu();
  const auto MSs = model.get_MSs();
  const auto MSc = model.get_MSc();
  const auto MSb = model.get_MSb();
  const auto MSt = model.get_MSt();
  const auto MGlu = model.get_MGlu();

  delta_alpha_s +=
      alpha_s / (2. * Pi) *
      (0.5 - 2 * FiniteLog(Abs(MGlu / currentScale)) -
       0.16666666666666666 * FiniteLog(Abs(MSd(0) / currentScale)) -
       0.16666666666666666 * FiniteLog(Abs(MSd(1) / currentScale)) -
       0.16666666666666666 * FiniteLog(Abs(MSu(0) / currentScale)) -
       0.16666666666666666 * FiniteLog(Abs(MSu(1) / currentScale)) -
       0.16666666666666666 * FiniteLog(Abs(MSs(0) / currentScale)) -
       0.16666666666666666 * FiniteLog(Abs(MSs(1) / currentScale)) -
       0.16666666666666666 * FiniteLog(Abs(MSc(0) / currentScale)) -
       0.16666666666666666 * FiniteLog(Abs(MSc(1) / currentScale)) -
       0.16666666666666666 * FiniteLog(Abs(MSb(0) / currentScale)) -
       0.16666666666666666 * FiniteLog(Abs(MSb(1) / currentScale)) -
       0.16666666666666666 * FiniteLog(Abs(MSt(0) / currentScale)) -
       0.16666666666666666 * FiniteLog(Abs(MSt(1) / currentScale)));

  return delta_alpha_s;
}

Eigen::Matrix<double, 3, 3> calculate_MFu_DRbar_tree_level(
    const NUHMSSMNoFVHimalaya_mass_eigenstates &model) {
  Eigen::Matrix<double, 3, 3> mf = ZEROMATRIX(3, 3);

  mf(0, 0) = model.get_MFu();
  mf(1, 1) = model.get_MFc();
  mf(2, 2) = model.get_MFt();

  return mf;
}

Eigen::Matrix<double, 3, 3> calculate_MFd_DRbar_tree_level(
    const NUHMSSMNoFVHimalaya_mass_eigenstates &model) {
  Eigen::Matrix<double, 3, 3> mf = ZEROMATRIX(3, 3);

  mf(0, 0) = model.get_MFd();
  mf(1, 1) = model.get_MFs();
  mf(2, 2) = model.get_MFb();

  return mf;
}

Eigen::Matrix<double, 3, 3> calculate_MFe_DRbar_tree_level(
    const NUHMSSMNoFVHimalaya_mass_eigenstates &model) {
  Eigen::Matrix<double, 3, 3> mf = ZEROMATRIX(3, 3);

  mf(0, 0) = model.get_MFe();
  mf(1, 1) = model.get_MFm();
  mf(2, 2) = model.get_MFtau();

  return mf;
}

double calculate_MFu_pole_1loop(int i,
                                const standard_model::Standard_model &sm_0l) {
  const double p = sm_0l.get_MFu(i);
  const auto self_energy_1 = Re(sm_0l.self_energy_Fu_1loop_1(p));
  const auto self_energy_PL = Re(sm_0l.self_energy_Fu_1loop_PL(p));
  const auto self_energy_PR = Re(sm_0l.self_energy_Fu_1loop_PR(p));
  const auto M_tree = sm_0l.get_mass_matrix_Fu();
  const auto M_loop = (M_tree - self_energy_PR * M_tree -
                       M_tree * self_energy_PL - self_energy_1)
                          .eval();

  Eigen::Array<double, 3, 1> MFu_pole;
  fs_svd(M_loop, MFu_pole);

  return MFu_pole(i);
}

double calculate_MFd_pole_1loop(int i,
                                const standard_model::Standard_model &sm_0l) {
  const double p = sm_0l.get_MFd(i);
  const auto self_energy_1 = Re(sm_0l.self_energy_Fd_1loop_1(p));
  const auto self_energy_PL = Re(sm_0l.self_energy_Fd_1loop_PL(p));
  const auto self_energy_PR = Re(sm_0l.self_energy_Fd_1loop_PR(p));
  const auto M_tree = sm_0l.get_mass_matrix_Fd();
  const auto M_loop = (M_tree - self_energy_PR * M_tree -
                       M_tree * self_energy_PL - self_energy_1)
                          .eval();

  Eigen::Array<double, 3, 1> MFd_pole;
  fs_svd(M_loop, MFd_pole);

  return MFd_pole(i);
}

double calculate_MFe_pole_1loop(int i,
                                const standard_model::Standard_model &sm_0l) {
  const auto p = sm_0l.get_MFe(i);
  const auto self_energy_1 = Re(sm_0l.self_energy_Fe_1loop_1(p));
  const auto self_energy_PL = Re(sm_0l.self_energy_Fe_1loop_PL(p));
  const auto self_energy_PR = Re(sm_0l.self_energy_Fe_1loop_PR(p));
  const auto M_tree = sm_0l.get_mass_matrix_Fe();
  const auto M_loop = (M_tree - self_energy_PR * M_tree -
                       M_tree * self_energy_PL - self_energy_1)
                          .eval();

  Eigen::Array<double, 3, 1> MFe_pole;
  fs_svd(M_loop, MFe_pole);

  return MFe_pole(i);
}

Eigen::Matrix<double, 3, 3>
calculate_MFu_pole_1loop(const standard_model::Standard_model &sm_0l) {
  Eigen::Matrix<double, 3, 1> M_pole;

#ifdef ENABLE_THREADS
  auto M_0 = global_thread_pool().run_packaged_task(
      [&sm_0l] { return calculate_MFu_pole_1loop(0, sm_0l); });
  auto M_1 = global_thread_pool().run_packaged_task(
      [&sm_0l] { return calculate_MFu_pole_1loop(1, sm_0l); });
  auto M_2 = global_thread_pool().run_packaged_task(
      [&sm_0l] { return calculate_MFu_pole_1loop(2, sm_0l); });
  M_pole << M_0.get(), M_1.get(), M_2.get();
#else
  M_pole << calculate_MFu_pole_1loop(0, sm_0l),
      calculate_MFu_pole_1loop(1, sm_0l), calculate_MFu_pole_1loop(2, sm_0l);
#endif

  return M_pole.asDiagonal();
}

Eigen::Matrix<double, 3, 3>
calculate_MFd_pole_1loop(const standard_model::Standard_model &sm_0l) {
  Eigen::Matrix<double, 3, 1> M_pole;

#ifdef ENABLE_THREADS
  auto M_0 = global_thread_pool().run_packaged_task(
      [&sm_0l] { return calculate_MFd_pole_1loop(0, sm_0l); });
  auto M_1 = global_thread_pool().run_packaged_task(
      [&sm_0l] { return calculate_MFd_pole_1loop(1, sm_0l); });
  auto M_2 = global_thread_pool().run_packaged_task(
      [&sm_0l] { return calculate_MFd_pole_1loop(2, sm_0l); });
  M_pole << M_0.get(), M_1.get(), M_2.get();
#else
  M_pole << calculate_MFd_pole_1loop(0, sm_0l),
      calculate_MFd_pole_1loop(1, sm_0l), calculate_MFd_pole_1loop(2, sm_0l);
#endif

  return M_pole.asDiagonal();
}

Eigen::Matrix<double, 3, 3>
calculate_MFe_pole_1loop(const standard_model::Standard_model &sm_0l) {
  Eigen::Matrix<double, 3, 1> M_pole;

#ifdef ENABLE_THREADS
  auto M_0 = global_thread_pool().run_packaged_task(
      [&sm_0l] { return calculate_MFe_pole_1loop(0, sm_0l); });
  auto M_1 = global_thread_pool().run_packaged_task(
      [&sm_0l] { return calculate_MFe_pole_1loop(1, sm_0l); });
  auto M_2 = global_thread_pool().run_packaged_task(
      [&sm_0l] { return calculate_MFe_pole_1loop(2, sm_0l); });
  M_pole << M_0.get(), M_1.get(), M_2.get();
#else
  M_pole << calculate_MFe_pole_1loop(0, sm_0l),
      calculate_MFe_pole_1loop(1, sm_0l), calculate_MFe_pole_1loop(2, sm_0l);
#endif

  return M_pole.asDiagonal();
}

double calculate_MFup_pole_1loop(
    const NUHMSSMNoFVHimalaya_mass_eigenstates &model_0l) {
  double m_pole = 0.;

  const double p = model_0l.get_MFu();
  const double self_energy_1 = Re(model_0l.self_energy_Fu_1loop_1(p));
  const double self_energy_PL = Re(model_0l.self_energy_Fu_1loop_PL(p));
  const double self_energy_PR = Re(model_0l.self_energy_Fu_1loop_PR(p));
  const double M_tree = model_0l.get_mass_matrix_Fu();
  const double M_loop = M_tree - self_energy_PR * M_tree -
                        M_tree * self_energy_PL - self_energy_1;

  m_pole = M_loop;

  return m_pole;
}

double calculate_MFdown_pole_1loop(
    const NUHMSSMNoFVHimalaya_mass_eigenstates &model_0l) {
  double m_pole = 0.;

  const double p = model_0l.get_MFd();
  const double self_energy_1 = Re(model_0l.self_energy_Fd_1loop_1(p));
  const double self_energy_PL = Re(model_0l.self_energy_Fd_1loop_PL(p));
  const double self_energy_PR = Re(model_0l.self_energy_Fd_1loop_PR(p));
  const double M_tree = model_0l.get_mass_matrix_Fd();
  const double M_loop = M_tree - self_energy_PR * M_tree -
                        M_tree * self_energy_PL - self_energy_1;

  m_pole = M_loop;

  return m_pole;
}

double calculate_MFele_pole_1loop(
    const NUHMSSMNoFVHimalaya_mass_eigenstates &model_0l) {
  double m_pole = 0.;

  const double p = model_0l.get_MFe();
  const double self_energy_1 = Re(model_0l.self_energy_Fe_1loop_1(p));
  const double self_energy_PL = Re(model_0l.self_energy_Fe_1loop_PL(p));
  const double self_energy_PR = Re(model_0l.self_energy_Fe_1loop_PR(p));
  const double M_tree = model_0l.get_mass_matrix_Fe();
  const double M_loop = M_tree - self_energy_PR * M_tree -
                        M_tree * self_energy_PL - self_energy_1;

  m_pole = M_loop;

  return m_pole;
}

double
calculate_MFc_pole_1loop(const NUHMSSMNoFVHimalaya_mass_eigenstates &model_0l) {
  double m_pole = 0.;

  const double p = model_0l.get_MFc();
  const double self_energy_1 = Re(model_0l.self_energy_Fc_1loop_1(p));
  const double self_energy_PL = Re(model_0l.self_energy_Fc_1loop_PL(p));
  const double self_energy_PR = Re(model_0l.self_energy_Fc_1loop_PR(p));
  const double M_tree = model_0l.get_mass_matrix_Fc();
  const double M_loop = M_tree - self_energy_PR * M_tree -
                        M_tree * self_energy_PL - self_energy_1;

  m_pole = M_loop;

  return m_pole;
}

double
calculate_MFs_pole_1loop(const NUHMSSMNoFVHimalaya_mass_eigenstates &model_0l) {
  double m_pole = 0.;

  const double p = model_0l.get_MFs();
  const double self_energy_1 =
      Re(model_0l.self_energy_Fs_1loop_1(p)); // TODO Fs_1loop
  const double self_energy_PL = Re(model_0l.self_energy_Fs_1loop_PL(p));
  const double self_energy_PR = Re(model_0l.self_energy_Fs_1loop_PR(p));
  const double M_tree = model_0l.get_mass_matrix_Fs();
  const double M_loop = M_tree - self_energy_PR * M_tree -
                        M_tree * self_energy_PL - self_energy_1;

  m_pole = M_loop;

  return m_pole;
}

double
calculate_MFm_pole_1loop(const NUHMSSMNoFVHimalaya_mass_eigenstates &model_0l) {
  double m_pole = 0.;

  const double p = model_0l.get_MFm();
  const double self_energy_1 = Re(model_0l.self_energy_Fm_1loop_1(p));
  const double self_energy_PL = Re(model_0l.self_energy_Fm_1loop_PL(p));
  const double self_energy_PR = Re(model_0l.self_energy_Fm_1loop_PR(p));
  const double M_tree = model_0l.get_mass_matrix_Fm();
  const double M_loop = M_tree - self_energy_PR * M_tree -
                        M_tree * self_energy_PL - self_energy_1;

  m_pole = M_loop;

  return m_pole;
}

double
calculate_MFt_pole_1loop(const NUHMSSMNoFVHimalaya_mass_eigenstates &model_0l) {
  double m_pole = 0.;

  const double p = model_0l.get_MFt();
  const double self_energy_1 = Re(model_0l.self_energy_Ft_1loop_1(p));
  const double self_energy_PL = Re(model_0l.self_energy_Ft_1loop_PL(p));
  const double self_energy_PR = Re(model_0l.self_energy_Ft_1loop_PR(p));
  const double M_tree = model_0l.get_mass_matrix_Ft();
  const double M_loop = M_tree - self_energy_PR * M_tree -
                        M_tree * self_energy_PL - self_energy_1;

  m_pole = M_loop;

  return m_pole;
}

double
calculate_MFb_pole_1loop(const NUHMSSMNoFVHimalaya_mass_eigenstates &model_0l) {
  double m_pole = 0.;

  const double p = model_0l.get_MFb();
  const double self_energy_1 = Re(model_0l.self_energy_Fb_1loop_1(p));
  const double self_energy_PL = Re(model_0l.self_energy_Fb_1loop_PL(p));
  const double self_energy_PR = Re(model_0l.self_energy_Fb_1loop_PR(p));
  const double M_tree = model_0l.get_mass_matrix_Fb();
  const double M_loop = M_tree - self_energy_PR * M_tree -
                        M_tree * self_energy_PL - self_energy_1;

  m_pole = M_loop;

  return m_pole;
}

double calculate_MFtau_pole_1loop(
    const NUHMSSMNoFVHimalaya_mass_eigenstates &model_0l) {
  double m_pole = 0.;

  const double p = model_0l.get_MFtau();
  const double self_energy_1 = Re(model_0l.self_energy_Ftau_1loop_1(p));
  const double self_energy_PL = Re(model_0l.self_energy_Ftau_1loop_PL(p));
  const double self_energy_PR = Re(model_0l.self_energy_Ftau_1loop_PR(p));
  const double M_tree = model_0l.get_mass_matrix_Ftau();
  const double M_loop = M_tree - self_energy_PR * M_tree -
                        M_tree * self_energy_PL - self_energy_1;

  m_pole = M_loop;

  return m_pole;
}

double
calculate_MFu_pole_1loop(int i,
                         const NUHMSSMNoFVHimalaya_mass_eigenstates &model) {
  switch (i) {
  case 0: {
    return calculate_MFup_pole_1loop(model);
  }

  case 1: {
    return calculate_MFc_pole_1loop(model);
  }

  case 2: {
    return calculate_MFt_pole_1loop(model);
  }
  }

  return calculate_MFt_pole_1loop(model);
}

double
calculate_MFd_pole_1loop(int i,
                         const NUHMSSMNoFVHimalaya_mass_eigenstates &model) {
  switch (i) {
  case 0: {
    return calculate_MFdown_pole_1loop(model);
  }

  case 1: {
    return calculate_MFs_pole_1loop(model);
  }

  case 2: {
    return calculate_MFb_pole_1loop(model);
  }
  }

  return calculate_MFb_pole_1loop(model);
}

double
calculate_MFe_pole_1loop(int i,
                         const NUHMSSMNoFVHimalaya_mass_eigenstates &model) {
  switch (i) {
  case 0: {
    return calculate_MFele_pole_1loop(model);
  }

  case 1: {
    return calculate_MFm_pole_1loop(model);
  }

  case 2: {
    return calculate_MFtau_pole_1loop(model);
  }
  }

  return calculate_MFtau_pole_1loop(model);
}

Eigen::Matrix<double, 3, 3>
calculate_MFu_pole_1loop(const NUHMSSMNoFVHimalaya_mass_eigenstates &model_0l) {

  Eigen::Matrix<double, 3, 1> M_pole;

#ifdef ENABLE_THREADS
  auto M_0 = global_thread_pool().run_packaged_task(
      [&model_0l] { return calculate_MFu_pole_1loop(0, model_0l); });
  auto M_1 = global_thread_pool().run_packaged_task(
      [&model_0l] { return calculate_MFu_pole_1loop(1, model_0l); });
  auto M_2 = global_thread_pool().run_packaged_task(
      [&model_0l] { return calculate_MFu_pole_1loop(2, model_0l); });
  M_pole << M_0.get(), M_1.get(), M_2.get();
#else
  M_pole << calculate_MFu_pole_1loop(0, model_0l),
      calculate_MFu_pole_1loop(1, model_0l),
      calculate_MFu_pole_1loop(2, model_0l);
#endif

  return M_pole.asDiagonal();
}

Eigen::Matrix<double, 3, 3>
calculate_MFd_pole_1loop(const NUHMSSMNoFVHimalaya_mass_eigenstates &model_0l) {
  Eigen::Matrix<double, 3, 1> M_pole;

#ifdef ENABLE_THREADS
  auto M_0 = global_thread_pool().run_packaged_task(
      [&model_0l] { return calculate_MFd_pole_1loop(0, model_0l); });
  auto M_1 = global_thread_pool().run_packaged_task(
      [&model_0l] { return calculate_MFd_pole_1loop(1, model_0l); });
  auto M_2 = global_thread_pool().run_packaged_task(
      [&model_0l] { return calculate_MFd_pole_1loop(2, model_0l); });
  M_pole << M_0.get(), M_1.get(), M_2.get();
#else
  M_pole << calculate_MFd_pole_1loop(0, model_0l),
      calculate_MFd_pole_1loop(1, model_0l),
      calculate_MFd_pole_1loop(2, model_0l);
#endif

  return M_pole.asDiagonal();
}

Eigen::Matrix<double, 3, 3>
calculate_MFe_pole_1loop(const NUHMSSMNoFVHimalaya_mass_eigenstates &model_0l) {
  Eigen::Matrix<double, 3, 1> M_pole;

#ifdef ENABLE_THREADS
  auto M_0 = global_thread_pool().run_packaged_task(
      [&model_0l] { return calculate_MFe_pole_1loop(0, model_0l); });
  auto M_1 = global_thread_pool().run_packaged_task(
      [&model_0l] { return calculate_MFe_pole_1loop(1, model_0l); });
  auto M_2 = global_thread_pool().run_packaged_task(
      [&model_0l] { return calculate_MFe_pole_1loop(2, model_0l); });
  M_pole << M_0.get(), M_1.get(), M_2.get();
#else
  M_pole << calculate_MFe_pole_1loop(0, model_0l),
      calculate_MFe_pole_1loop(1, model_0l),
      calculate_MFe_pole_1loop(2, model_0l);
#endif

  return M_pole.asDiagonal();
}

Eigen::Matrix<double, 3, 3>
calculate_MFu_DRbar_1loop(const standard_model::Standard_model &sm_0l,
                          const NUHMSSMNoFVHimalaya_mass_eigenstates &model) {
  Eigen::Matrix<double, 3, 3> mf_sm_1l = ZEROMATRIX(3, 3);

  const auto Mf_sm = calculate_MFu_pole_1loop(sm_0l);
  const auto Mf_bsm = calculate_MFu_pole_1loop(model);
  const auto mf_sm_0l =
      calculate_MFu_DRbar_tree_level(model); // treelevel matching

  mf_sm_1l = Mf_bsm - Mf_sm + mf_sm_0l;

  return Abs(mf_sm_1l);
}

Eigen::Matrix<double, 3, 3>
calculate_MFd_DRbar_1loop(const standard_model::Standard_model &sm_0l,
                          const NUHMSSMNoFVHimalaya_mass_eigenstates &model) {
  Eigen::Matrix<double, 3, 3> mf_sm_1l = ZEROMATRIX(3, 3);

  const auto Mf_sm = calculate_MFd_pole_1loop(sm_0l);
  const auto Mf_bsm = calculate_MFd_pole_1loop(model);
  const auto mf_sm_0l =
      calculate_MFd_DRbar_tree_level(model); // treelevel matching

  mf_sm_1l = Mf_bsm - Mf_sm + mf_sm_0l;

  return Abs(mf_sm_1l);
}

Eigen::Matrix<double, 3, 3>
calculate_MFe_DRbar_1loop(const standard_model::Standard_model &sm_0l,
                          const NUHMSSMNoFVHimalaya_mass_eigenstates &model) {
  Eigen::Matrix<double, 3, 3> mf_sm_1l = ZEROMATRIX(3, 3);

  const auto Mf_sm = calculate_MFe_pole_1loop(sm_0l);
  const auto Mf_bsm = calculate_MFe_pole_1loop(model);
  const auto mf_sm_0l =
      calculate_MFe_DRbar_tree_level(model); // treelevel matching

  mf_sm_1l = Mf_bsm - Mf_sm + mf_sm_0l;

  return Abs(mf_sm_1l);
}

double calculate_MFt_MSbar_2loop(
    const standard_model::Standard_model &sm_0l,
    const NUHMSSMNoFVHimalaya_mass_eigenstates &model_0l) {
  auto model = model_0l;
  auto sm = sm_0l;

  sm.set_pole_mass_loop_order(2);
  sm.set_top_QCD_order(1);
  model.set_pole_mass_loop_order(2);
  model.set_top_QCD_order(1);

  double mst_1, mst_2, theta_t;
  model.calculate_MTopSquark_3rd_generation(mst_1, mst_2, theta_t);

  double Q2 = Sqr(model.get_scale());
  double mt = model.get_MFt();
  double mt2 = Sqr(mt);
  double gs = model.get_g3();
  double gs2 = Sqr(model.get_g3());
  double alpha_s = gs2 / (4. * Pi);

  double k = 1. / Power(4 * Pi, 2);
  double logmt = Log(mt2 / Q2);
  double yt_SM = sm.get_Yu(2, 2);

  mssm_twoloop_mt::Parameters pars;
  pars.g3 = gs;
  pars.mt = mt;
  pars.mg = model.get_MGlu();
  pars.mst1 = mst_1;
  pars.mst2 = mst_2;
  pars.msusy = Sqrt(Sqrt(Abs(model.get_md2(2, 2)))) *
               Sqrt(Sqrt(Abs(model.get_mq2(2, 2))));
  pars.xt = Sin(2 * theta_t) * (Sqr(mst_1) - Sqr(mst_2)) / (2. * pars.mt);
  pars.Q = model.get_scale();

  double delta_alpha_s = calculate_delta_alpha_s(alpha_s, model_0l);

  // top mass contribution at O(as) in MSSM and SM
  double sqcd_1l_over_mt = mssm_twoloop_mt::dMt_over_mt_1loop(pars);
  double qcd_1l_SM_over_mt =
      0.008443431970194815 * (4. - 3. * Log(mt2 / Q2)) * gs2;

  // top mass 1-loop contribution at O(as) SM subscript
  double qcd1l_mt = -4. / 3. * k * gs2 * mt * (2. + 3. * logmt);
  double qcd1l_gs = 8. / 3. * k * gs2 * mt * (4. - 3. * logmt);

  double delta_mt_over_mt = sqcd_1l_over_mt - qcd_1l_SM_over_mt;
  double delta_gs_over_gs = -0.5 * delta_alpha_s;

  double mfcon = delta_mt_over_mt * qcd1l_mt + delta_gs_over_gs * qcd1l_gs;

  model.calculate_MFt_pole();
  sm.calculate_MFu_pole();

  const auto Mf_bsm = model.get_physical().MFt;
  const auto Mf_sm = sm.get_physical().MFu(2);
  const auto mf_sm_0l = sm.get_MFu(2);

  double mf_sm = Mf_bsm - Mf_sm + mf_sm_0l - mfcon;

  return Abs(mf_sm);
}

/**
 * Calculates tadpoles over vevs (at given fixed loop order)
 *
 * @param model model parameters
 * @param loop_order loop order
 *
 * @return tadpole at loop order
 *
 */
Eigen::Matrix<double, 2, 1>
calculate_tadpole_over_vevs(NUHMSSMNoFVHimalaya_mass_eigenstates model,
                            int loop_order) {
  if (loop_order == 0) {
    model.set_ewsb_loop_order(loop_order);
    return model.tadpole_equations_over_vevs();
  }

  model.set_ewsb_loop_order(loop_order);
  const auto tadpole_lo = model.tadpole_equations_over_vevs(); // nL

  model.set_ewsb_loop_order(loop_order - 1);
  const auto tadpole_lom1 = model.tadpole_equations_over_vevs(); // (n-1)L

  return (tadpole_lo - tadpole_lom1)
      .eval(); // returns just the difference of tadpoles
}

/**
 * Calculates SM pole mass at 1l.
 */
double calculate_Mh2_pole_1l(const standard_model::Standard_model &sm_0l) {
  const double p = sm_0l.get_Mhh();
  const double self_energy = Re(sm_0l.self_energy_hh_1loop(p));
  const double tadpole = Re(sm_0l.tadpole_hh_1loop() / sm_0l.get_v());
  const double mh2_tree = Sqr(sm_0l.get_Mhh());
  const double Mh2_pole = mh2_tree - self_energy + tadpole;

  return Mh2_pole;
}

/**
 * Calculates tree-level Higgs mass matrix in the NUHMSSMNoFVHimalaya without
 * including tadpoles implicitly.  This is achieved by solving the
 * EWSB equations at tree-level in order to avoid the inclusion of
 * loop tadpoles.
 *
 * @param model NUHMSSMNoFVHimalaya parameters
 * @return tree-level Higgs mass matrix in the NUHMSSMNoFVHimalaya w/o tadpoles
 */
auto calculate_mh2_tree_level(NUHMSSMNoFVHimalaya_mass_eigenstates model)
    -> decltype(model.get_mass_matrix_hh()) {
  model.solve_ewsb_tree_level();
  return model.get_mass_matrix_hh();
}

/**
 * Calculates BSM pole mass at 1l.
 */

double calculate_Mh2_pole_1l(const NUHMSSMNoFVHimalaya_mass_eigenstates &model,
                             int idx) {

  // calculate tree-level mass matrix
  const auto mh2_tree = calculate_mh2_tree_level(model);

  // calculate 1L self-energy using tree-level parameters
  const double p = model.get_Mhh(idx);
  const auto self_energy = Re(model.self_energy_hh_1loop(p));

  // calculate 1L tadpoles using tree-level parameters
  const auto tadpole = calculate_tadpole_over_vevs(model, 1);

  double Mh2_pole = 0.;

  const auto M_loop = (mh2_tree - self_energy -
                       Eigen::Matrix<double, 2, 2>(tadpole.asDiagonal()))
                          .eval();

  Eigen::Array<double, 2, 1> M_pole;
  fs_diagonalize_hermitian(M_loop, M_pole);

  Mh2_pole = M_pole(idx);

  return Mh2_pole;
}

double
calculate_Mh2_shift_1l_2l(const NUHMSSMNoFVHimalaya_mass_eigenstates &model,
                          int idx) {

  auto model_1l = model;

  int nrows = model.get_Yu().rows();
  int ncols = model.get_Yu().cols();

  for (int i = 0; i < nrows; i++) {
    for (int k = 0; k < ncols; k++) {

      if (model.HIGGS_2LOOP_CORRECTION_AT_AT && i == 2 && k == 2) {
        continue;
      }
      model_1l.set_Yu(i, k, .0);
      model_1l.set_TYu(i, k, .0);
      model_1l.set_Yd(i, k, .0);
      model_1l.set_TYd(i, k, .0);
    }
  }

  nrows = model.get_Ye().rows();
  ncols = model.get_Ye().cols();

  for (int i = 0; i < nrows; i++) {
    for (int k = 0; k < ncols; k++) {

      if (model.HIGGS_2LOOP_CORRECTION_ATAU_ATAU && i == 2 && k == 2) {
        continue;
      }
      model_1l.set_Ye(i, k, .0);
      model_1l.set_TYe(i, k, .0);
    }
  }

  model_1l.calculate_DRbar_masses();

  const double mh2_tree = Sqr(model_1l.get_Mhh(idx));
  const auto mh2_shift_1l = calculate_Mh2_pole_1l(model_1l, idx) - mh2_tree;

  return mh2_shift_1l;
}

double calculate_Mh2_shift_1l_2l(const standard_model::Standard_model &sm) {
  using namespace flexiblesusy::sm_twoloophiggs;

  auto sm_1l = sm;

  int nrows = sm.get_Yu().rows();
  int ncols = sm.get_Yu().cols();

  for (int i = 0; i < nrows; i++) {
    for (int k = 0; k < ncols; k++) {

      if (sm.HIGGS_2LOOP_CORRECTION_AT_AT && i == 2 && k == 2) {
        continue;
      }
      sm_1l.set_Yu(i, k, .0);
      sm_1l.set_Yd(i, k, .0);
    }
  }

  nrows = sm.get_Ye().rows();
  ncols = sm.get_Ye().cols();

  for (int i = 0; i < nrows; i++) {
    for (int k = 0; k < ncols; k++) {

      if (sm.HIGGS_2LOOP_CORRECTION_ATAU_ATAU && i == 2 && k == 2) {
        continue;
      }
      sm_1l.set_Ye(i, k, .0);
    }
  }

  sm_1l.calculate_DRbar_masses();

  const double mh2_tree = Sqr(sm.get_Mhh());
  const double mh2_shift_1l = calculate_Mh2_pole_1l(sm_1l) - mh2_tree;

  return mh2_shift_1l;
}

double
calculate_mt_shift_2L(const NUHMSSMNoFVHimalaya_mass_eigenstates &model_norm,
                      const standard_model::Standard_model &sm) {
  auto model = model_norm;
  auto sm_1l = sm;

  double g3oPMS2 = Sqr(model.get_g3() / Pi);
  double g3oPMS4 = Sqr(g3oPMS2);
  double mtDR = model.get_MFt();

  double mt_shift_2l = mtDR * 59. * g3oPMS4 / 2304.;

  return mt_shift_2l;
}

/**
 * incomplete higher-order contributions to estimate the high-scale
 * uncertainty
 */

double
calculate_delta_lam_3l(const NUHMSSMNoFVHimalaya_mass_eigenstates &model_norm,
                       const standard_model::Standard_model &sm_0l,
                       const standard_model::Standard_model &sm_1l,
                       const standard_model::Standard_model &sm_1lGl) {

  auto model = model_norm;
  auto sm0l = sm_0l;
  auto sm1l = sm_1l;
  auto sm1lGl = sm_1lGl;

  double mh = model.get_Mhh(0);
  double yt = model.get_Yu(2, 2);
  double ytau = model.get_Ye(2, 2);
  double mt = model.get_MFt();
  double gs = model.get_g3();
  double gysm = sm0l.get_g1() * standard_model_info::normalization_g1;
  double gwsm = sm0l.get_g2() * standard_model_info::normalization_g2;
  double ytsm = sm0l.get_Yu(2, 2);
  double gssm = sm0l.get_Yu(2, 2);
  double sb = model.get_vu() / Sqrt(Sqr(model.get_vu()) + Sqr(model.get_vd()));
  double cb = model.get_vd() / Sqrt(Sqr(model.get_vu()) + Sqr(model.get_vd()));
  double b = ArcCos(cb);
  double MS = Sqrt(model.get_MSt(0) * model.get_MSt(1));
  double Xhat =
      (model.get_TYu(2, 2) / (model.get_Yu(2, 2)) - model.get_Mu() * cb / sb) /
      MS;

  double Xhat2 = Xhat * Xhat;
  double tb = sb / cb;
  double k = 1. / Power(4 * Pi, 2);
  double K = -0.1953256;
  double v = Sqrt(Sqr(model.get_vu()) + Sqr(model.get_vd()));
  double lam = sm0l.get_Lambdax();

  double delta_lam_tau_top = 2. * Power(k, 2) * Power(yt, 4) * Power(ytau, 2) *
                             (Power(sb, 4)) * Xhat2 * (-1. + Xhat2);

  double delta_lam_3l4l =
      -(Power(k, 3) * Power(yt, 6) *
        (16 * Power(gs, 2) *
             (-18 + 6 * Power(Pi, 2) * (-1 + Xhat) - 30 * Xhat +
              96 * Power(Xhat, 2) - 64 * Power(Xhat, 3) - 7 * Power(Xhat, 4) +
              5 * Power(Xhat, 5)) -
         3 *
             (-96 + 140 * Power(Xhat, 2) + 106 * Power(Xhat, 4) -
              18 * Power(Xhat, 6) + Power(Xhat, 8) +
              Power(Pi, 2) * (8 - 6 * Power(Xhat, 2)) +
              192 * K * (2 - 3 * Power(Xhat, 2) + Power(Xhat, 4))) *
             Power(yt, 2))) /
      12.;

  double del_lam_rep_resum =
      -(Power(gssm, 2) * Power(k, 3) * Power(Xhat, 7) * Power(ytsm, 6) *
        (83 + 288 * K + 9 * (5 + 32 * K) * Cos(2 * b)) * Power(Csc(b), 2)) /
      3.;
  double del_lam_rep =
      -(Power(k, 3) * Xhat * Power(ytsm, 4) * Power(Csc(b), 2) *
        (108 * Power(Xhat, 7) * Power(ytsm, 4) *
             (26 + 96 * K + 3 * (5 + 32 * K) * Cos(2 * b)) +
         432 * (19 + 96 * K) * Power(Xhat, 6) * Power(ytsm, 4) * Cot(b) +
         27648 * Power(gssm, 2) * Power(ytsm, 2) * Power(Cot(b), 2) -
         (9 * Xhat * Power(ytsm, 2) *
          (10880 * Power(gssm, 2) - 16896 * Power(gssm, 2) * K +
           855 * Power(ytsm, 2) + 4752 * K * Power(ytsm, 2) +
           (224 * Power(gssm, 2) * (-79 + 120 * K) +
            9 * (323 + 840 * K) * Power(ytsm, 2)) *
               Cos(2 * b) +
           (-384 * Power(gssm, 2) * (-17 + 52 * K) +
            27 * (79 + 208 * K) * Power(ytsm, 2)) *
               Cos(4 * b) +
           288 * Power(gssm, 2) * Cos(6 * b) +
           9984 * Power(gssm, 2) * K * Cos(6 * b) +
           81 * Power(ytsm, 2) * Cos(6 * b) +
           2808 * K * Power(ytsm, 2) * Cos(6 * b)) *
          Power(Csc(b), 4)) /
             2. -
         Power(Xhat, 3) *
             (36864 * Power(gssm, 2) * (5 + 24 * K) * Power(ytsm, 2) * Cot(b) -
              9477 * Power(ytsm, 4) * Power(Cot(b), 4) +
              4 *
                  (4096 * Power(gssm, 4) -
                   11088 * Power(gssm, 2) * Power(ytsm, 2) +
                   10368 * Power(gssm, 2) * K * Power(ytsm, 2) -
                   4995 * Power(ytsm, 4) + 7128 * K * Power(ytsm, 4) +
                   (-4096 * Power(gssm, 4) +
                    6768 * Power(gssm, 2) * Power(ytsm, 2) +
                    2592 * Power(ytsm, 4)) *
                       Cos(2 * b) +
                   9 * Power(ytsm, 2) *
                       (32 * Power(gssm, 2) * (7 + 60 * K) +
                        9 * (3 + 104 * K) * Power(ytsm, 2)) *
                       Cos(4 * b)) *
                  Power(Csc(b), 2) +
              36 * Power(Cot(b), 2) *
                  (944 * Power(gssm, 2) * Power(ytsm, 2) +
                   1143 * Power(ytsm, 4) -
                   9 * Power(ytsm, 4) *
                       (-15 + 36 * K + (7 + 60 * K) * Cos(4 * b)) *
                       Power(Csc(b), 2))) +
         64 * Power(Xhat, 2) *
             (576 * Power(gssm, 2) * (5 + 24 * K) * Power(ytsm, 2) * Cot(b) -
              2556 * Power(gssm, 2) * Power(ytsm, 2) * Power(Cot(b), 2) -
              162 * (5 + 24 * K) * Power(ytsm, 4) * Power(Cot(b), 3) +
              Power(gssm, 2) *
                  (32 * (13 * Power(gssm, 2) + 63 * Power(ytsm, 2)) -
                   9 * Power(ytsm, 2) *
                       (-89 + 264 * K + (9 + 312 * K) * Cos(4 * b)) *
                       Power(Csc(b), 2))) +
         6 * Power(Xhat, 4) * Power(ytsm, 2) * Power(Csc(b), 3) *
             ((-96 * Power(gssm, 2) * (19 + 96 * K) +
               9 * (11 + 96 * K) * Power(ytsm, 2)) *
                  Cos(b) +
              3 *
                  (32 * Power(gssm, 2) * (19 + 96 * K) +
                   3 * (217 + 1056 * K) * Power(ytsm, 2)) *
                  Cos(3 * b) +
              16 * Power(gssm, 2) *
                  (-343 + 432 * K + 529 * Cos(2 * b) +
                   12 * (7 + 60 * K) * Cos(4 * b)) *
                  Sin(b)) +
         (3 * Power(Xhat, 5) * Power(ytsm, 2) * Power(Csc(b), 2) *
          (-3616 * Power(gssm, 2) - 9216 * Power(gssm, 2) * K -
           1485 * Power(ytsm, 2) + 12960 * K * Power(ytsm, 2) +
           8 * (272 * Power(gssm, 2) + 27 * (35 + 48 * K) * Power(ytsm, 2)) *
               Cos(2 * b) +
           9 * (32 * Power(gssm, 2) * (5 + 32 * K) + (157 + 1248 * K) * Power(ytsm, 2)) *
               Cos(4 * b) +
           14592 * Power(gssm, 2) * Sin(2 * b) +
           73728 * Power(gssm, 2) * K * Sin(2 * b))) /
             2.)) /
      576.;

  double vderiv =
      -(k *
        (9 * Power(gwsm, 6) * (18 + Sqrt(3) * Pi) +
         3 * Power(gysm, 6) * (18 + Sqrt(3) * Pi) -
         48 * Power(gwsm, 5) * Sqrt(lam) * (9 + 2 * Sqrt(3) * Pi) +
         64 * Power(gwsm, 3) * Power(lam, 1.5) * (9 + 2 * Sqrt(3) * Pi) -
         64 * gwsm * Power(lam, 2.5) * (9 + 2 * Sqrt(3) * Pi) +
         4 * Power(gysm, 2) * Power(lam, 1.5) *
             (-9 * Sqrt(lam) * (12 + Sqrt(3) * Pi) +
              8 * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) *
                  (9 + 2 * Sqrt(3) * Pi)) +
         Power(gysm, 4) * (-24 * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) *
                               Sqrt(lam) * (9 + 2 * Sqrt(3) * Pi) +
                           2 * lam * (99 + 13 * Sqrt(3) * Pi)) +
         3 * Power(gwsm, 4) *
             (3 * Power(gysm, 2) * (18 + Sqrt(3) * Pi) -
              8 * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) * Sqrt(lam) *
                  (9 + 2 * Sqrt(3) * Pi) +
              2 * lam * (99 + 13 * Sqrt(3) * Pi)) +
         Power(gwsm, 2) *
             (-108 * Power(lam, 2) * (12 + Sqrt(3) * Pi) +
              9 * Power(gysm, 4) * (18 + Sqrt(3) * Pi) +
              32 * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) * Power(lam, 1.5) *
                  (9 + 2 * Sqrt(3) * Pi) +
              Power(gysm, 2) * (-48 * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) *
                                    Sqrt(lam) * (9 + 2 * Sqrt(3) * Pi) +
                                4 * lam * (99 + 13 * Sqrt(3) * Pi))) -
         16 * lam *
             (2 * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) * Power(lam, 1.5) *
                  (9 + 2 * Sqrt(3) * Pi) +
              3 * Power(lam, 2) * (-81 + 11 * Sqrt(3) * Pi) +
              135 * lam * Power(ytsm, 2) - 162 * Power(ytsm, 4)))) /
      (216. * lam * v);

  double vderiv_at =
      -(k *
        (9 * Power(gwsm, 6) * (18 + Sqrt(3) * Pi) +
         3 * Power(gysm, 6) * (18 + Sqrt(3) * Pi) -
         48 * Power(gwsm, 5) * Sqrt(lam) * (9 + 2 * Sqrt(3) * Pi) +
         64 * Power(gwsm, 3) * Power(lam, 1.5) * (9 + 2 * Sqrt(3) * Pi) -
         64 * gwsm * Power(lam, 2.5) * (9 + 2 * Sqrt(3) * Pi) +
         4 * Power(gysm, 2) * Power(lam, 1.5) *
             (-9 * Sqrt(lam) * (12 + Sqrt(3) * Pi) +
              8 * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) *
                  (9 + 2 * Sqrt(3) * Pi)) +
         Power(gysm, 4) * (-24 * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) *
                               Sqrt(lam) * (9 + 2 * Sqrt(3) * Pi) +
                           2 * lam * (99 + 13 * Sqrt(3) * Pi)) +
         3 * Power(gwsm, 4) *
             (3 * Power(gysm, 2) * (18 + Sqrt(3) * Pi) -
              8 * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) * Sqrt(lam) *
                  (9 + 2 * Sqrt(3) * Pi) +
              2 * lam * (99 + 13 * Sqrt(3) * Pi)) +
         Power(gwsm, 2) *
             (-108 * Power(lam, 2) * (12 + Sqrt(3) * Pi) +
              9 * Power(gysm, 4) * (18 + Sqrt(3) * Pi) +
              32 * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) * Power(lam, 1.5) *
                  (9 + 2 * Sqrt(3) * Pi) +
              Power(gysm, 2) * (-48 * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) *
                                    Sqrt(lam) * (9 + 2 * Sqrt(3) * Pi) +
                                4 * lam * (99 + 13 * Sqrt(3) * Pi))) -
         16 * Power(lam, 2) *
             (2 * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) * Sqrt(lam) *
                  (9 + 2 * Sqrt(3) * Pi) +
              3 * lam * (-81 + 11 * Sqrt(3) * Pi) + 135 * Power(ytsm, 2)))) /
      (216. * lam * v);

  double ytderiv = 2 * k * ytsm * (5 * lam - 6 * Power(ytsm, 2));

  double ytderiv_at = 10 * k * lam * ytsm;

  double lamderiv =
      (k *
       (162 * Power(gysm, 4) * lam + 9 * Power(gwsm, 6) * (18 + Sqrt(3) * Pi) +
        3 * Power(gysm, 6) * (18 + Sqrt(3) * Pi) -
        24 * Power(gwsm, 5) * Sqrt(lam) * (9 + 2 * Sqrt(3) * Pi) -
        12 * Power(gysm, 4) * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) *
            Sqrt(lam) * (9 + 2 * Sqrt(3) * Pi) -
        32 * Power(gwsm, 3) * Power(lam, 1.5) * (9 + 2 * Sqrt(3) * Pi) +
        96 * gwsm * Power(lam, 2.5) * (9 + 2 * Sqrt(3) * Pi) +
        4 * Power(gysm, 2) * Power(lam, 1.5) *
            (9 * Sqrt(lam) * (12 + Sqrt(3) * Pi) -
             4 * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) *
                 (9 + 2 * Sqrt(3) * Pi)) +
        3 * Power(gwsm, 4) *
            (162 * lam + 3 * Power(gysm, 2) * (18 + Sqrt(3) * Pi) -
             4 * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) * Sqrt(lam) *
                 (9 + 2 * Sqrt(3) * Pi)) +
        Power(gwsm, 2) *
            (324 * Power(gysm, 2) * lam +
             108 * Power(lam, 2) * (12 + Sqrt(3) * Pi) +
             9 * Power(gysm, 4) * (18 + Sqrt(3) * Pi) -
             24 * Power(gysm, 2) * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) *
                 Sqrt(lam) * (9 + 2 * Sqrt(3) * Pi) -
             16 * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) * Power(lam, 1.5) *
                 (9 + 2 * Sqrt(3) * Pi)) +
        48 * Power(lam, 2) *
            (Sqrt(Power(gwsm, 2) + Power(gysm, 2)) * Sqrt(lam) *
                 (9 + 2 * Sqrt(3) * Pi) +
             lam * (-216 + 22 * Sqrt(3) * Pi) + 18 * Power(ytsm, 2)))) /
      (432. * Power(lam, 2));

  double lamderiv_at =
      (k *
       (162 * Power(gysm, 4) * lam + 9 * Power(gwsm, 6) * (18 + Sqrt(3) * Pi) +
        3 * Power(gysm, 6) * (18 + Sqrt(3) * Pi) -
        24 * Power(gwsm, 5) * Sqrt(lam) * (9 + 2 * Sqrt(3) * Pi) -
        12 * Power(gysm, 4) * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) *
            Sqrt(lam) * (9 + 2 * Sqrt(3) * Pi) -
        32 * Power(gwsm, 3) * Power(lam, 1.5) * (9 + 2 * Sqrt(3) * Pi) +
        96 * gwsm * Power(lam, 2.5) * (9 + 2 * Sqrt(3) * Pi) +
        48 * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) * Power(lam, 2.5) *
            (9 + 2 * Sqrt(3) * Pi) +
        96 * Power(lam, 3) * (-108 + 11 * Sqrt(3) * Pi) +
        4 * Power(gysm, 2) * Power(lam, 1.5) *
            (9 * Sqrt(lam) * (12 + Sqrt(3) * Pi) -
             4 * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) *
                 (9 + 2 * Sqrt(3) * Pi)) +
        3 * Power(gwsm, 4) *
            (162 * lam + 3 * Power(gysm, 2) * (18 + Sqrt(3) * Pi) -
             4 * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) * Sqrt(lam) *
                 (9 + 2 * Sqrt(3) * Pi)) +
        Power(gwsm, 2) *
            (324 * Power(gysm, 2) * lam +
             108 * Power(lam, 2) * (12 + Sqrt(3) * Pi) +
             9 * Power(gysm, 4) * (18 + Sqrt(3) * Pi) -
             24 * Power(gysm, 2) * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) *
                 Sqrt(lam) * (9 + 2 * Sqrt(3) * Pi) -
             16 * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) * Power(lam, 1.5) *
                 (9 + 2 * Sqrt(3) * Pi)))) /
      (432. * Power(lam, 2));

  double g2deriv =
      -(k *
        (-120 * Power(gwsm, 4) * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) *
             Sqrt(lam) * (9 + 2 * Sqrt(3) * Pi) +
         96 * Power(gwsm, 2) * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) *
             Power(lam, 1.5) * (9 + 2 * Sqrt(3) * Pi) -
         32 * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) * Power(lam, 2.5) *
             (9 + 2 * Sqrt(3) * Pi) +
         3 * Power(gwsm, 5) *
             (9 * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) * (18 + Sqrt(3) * Pi) -
              20 * Sqrt(lam) * (9 + 2 * Sqrt(3) * Pi)) +
         6 * Power(gwsm, 3) *
             (8 * Power(lam, 1.5) * (9 + 2 * Sqrt(3) * Pi) +
              2 * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) * lam *
                  (180 + 13 * Sqrt(3) * Pi) +
              Power(gysm, 2) * (3 * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) *
                                    (18 + Sqrt(3) * Pi) -
                                20 * Sqrt(lam) * (9 + 2 * Sqrt(3) * Pi))) +
         gwsm *
             (3 * Power(gysm, 4) *
                  (3 * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) *
                       (18 + Sqrt(3) * Pi) -
                   20 * Sqrt(lam) * (9 + 2 * Sqrt(3) * Pi)) -
              4 * Power(lam, 2) *
                  (27 * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) *
                       (18 + Sqrt(3) * Pi) +
                   4 * Sqrt(lam) * (9 + 2 * Sqrt(3) * Pi)) +
              Power(gysm, 2) * (48 * Power(lam, 1.5) * (9 + 2 * Sqrt(3) * Pi) +
                                4 * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) *
                                    lam * (180 + 13 * Sqrt(3) * Pi))))) /
      (216. * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) * lam);

  double gyderiv =
      (gysm * k *
       (216 * (Power(gwsm, 2) + Power(gysm, 2)) -
        ((3 * Power(gwsm, 4) + 3 * Power(gysm, 4) +
          Power(gwsm, 2) * (6 * Power(gysm, 2) - 4 * lam) -
          4 * Power(gysm, 2) * lam + 4 * Power(lam, 2)) *
         (Sqrt(Power(gwsm, 2) + Power(gysm, 2)) * (18 + Sqrt(3) * Pi) -
          4 * Sqrt(lam) * (9 + 2 * Sqrt(3) * Pi))) /
            (Sqrt(Power(gwsm, 2) + Power(gysm, 2)) * lam) -
        (2 * (3 * Power(gwsm, 2) + 3 * Power(gysm, 2) - 2 * lam) *
         (Power(gwsm, 2) * (18 + Sqrt(3) * Pi) +
          Power(gysm, 2) * (18 + Sqrt(3) * Pi) +
          10 * lam * (18 + Sqrt(3) * Pi) -
          8 * Sqrt(Power(gwsm, 2) + Power(gysm, 2)) * Sqrt(lam) *
              (9 + 2 * Sqrt(3) * Pi))) /
            lam)) /
      216.;

  double delta_lam_gauge_2l =
      -lamderiv_at * (sm1lGl.get_Lambdax() - lam) -
      lamderiv * (sm1l.get_Lambdax() - sm1lGl.get_Lambdax()) // Important
      - ytderiv_at * (sm1lGl.get_Yu(2, 2) - ytsm) -
      ytderiv * (sm1l.get_Yu(2, 2) - sm1lGl.get_Yu(2, 2)) -
      vderiv_at * (sm1lGl.get_v() - v) -
      vderiv * (sm1l.get_v() - sm1lGl.get_v()) -
      g2deriv * (sm1l.get_g2() * standard_model_info::normalization_g2 - gwsm) -
      gyderiv * (sm1l.get_g1() * standard_model_info::normalization_g1 - gysm) +
      3 * Sqr(mh) / Power(v, 4) * Sqr(sm1l.get_v() - v);

  return -del_lam_rep -
         del_lam_rep_resum; //+delta_lam_3l4l; //-del_lam_rep ;//
                            // delta_lam_gauge_2l;//- del_lam_rep_resum; //
}

double
calculate_delta_lam_2l(const NUHMSSMNoFVHimalaya_mass_eigenstates &model_norm,
                       const standard_model::Standard_model &sm_0l,
                       const standard_model::Standard_model &sm_1l,
                       const standard_model::Standard_model &sm_1lGl) {

  auto model = model_norm;
  auto sm0l = sm_0l;
  auto sm1l = sm_1l;
  double yt = model.get_Yu(2, 2);
  double gs = model.get_g3();
  double sb = model.get_vu() / Sqrt(Sqr(model.get_vu()) + Sqr(model.get_vd()));
  double cb = model.get_vd() / Sqrt(Sqr(model.get_vu()) + Sqr(model.get_vd()));
  double MS = Sqrt(model.get_MSt(0) * model.get_MSt(1));
  double Xhat =
      (model.get_TYu(2, 2) / (model.get_Yu(2, 2)) - model.get_Mu() * cb / sb) /
      MS;
  double k = 1. / Power(4 * Pi, 2);

  double delta_lam =
      (2 * Power(gs, 4) * Power(k, 3) *
       (2243 - 1312 * Xhat + 336 * Power(Xhat, 2)) * Power(yt, 4)) /
      9.;

  return delta_lam + calculate_delta_lam_3l(model, sm0l, sm1l, sm_1lGl);
}

double
calculate_Mh2_shift_2l(const NUHMSSMNoFVHimalaya_mass_eigenstates &model_gl,
                       int idx) {
  auto model = model_gl;
  const double p = 0.;
  const auto mh2_shift_1l = calculate_Mh2_shift_1l_2l(model, idx);

  // calculate 2L self-energy using tree-level parameters
  auto self_energy_2l = Re(model.self_energy_hh_2loop());
  const auto tadpole_2l = calculate_tadpole_over_vevs(model, 2);

  int nrows = model.get_Yu().rows();
  int ncols = model.get_Yu().cols();

  for (int i = 0; i < nrows; i++) {
    for (int k = 0; k < ncols; k++) {

      if (model.HIGGS_2LOOP_CORRECTION_AT_AT && i == 2 && k == 2) {
        continue;
      }
      model.set_Yu(i, k, .0);
      model.set_TYu(i, k, .0);
      model.set_Yd(i, k, .0);
      model.set_TYd(i, k, .0);
    }
  }

  nrows = model.get_Ye().rows();
  ncols = model.get_Ye().cols();

  for (int i = 0; i < nrows; i++) {
    for (int k = 0; k < ncols; k++) {

      if (model.HIGGS_2LOOP_CORRECTION_ATAU_ATAU && i == 2 && k == 2) {
        continue;
      }
      model.set_Ye(i, k, .0);
      model.set_TYe(i, k, .0);
    }
  }

  model.calculate_DRbar_masses();

  self_energy_2l += Re(model.self_energy_hh_1loop_deriv_p2(p) * mh2_shift_1l);
  // calculate 2L tadpoles using tree-level parameters

  const auto Mh2_loop = (calculate_mh2_tree_level(model) - self_energy_2l -
                         Eigen::Matrix<double, 2, 2>(tadpole_2l.asDiagonal()))
                            .eval();

  Eigen::Array<double, 2, 1> Mh2_pole;
  fs_diagonalize_hermitian(Mh2_loop, Mh2_pole);
  double mh2_shift = Mh2_pole(idx) - Sqr(model.get_Mhh(idx));

  return mh2_shift;
}

double calculate_Mh2_shift_2l(const standard_model::Standard_model &sm,
                              const NUHMSSMNoFVHimalaya_mass_eigenstates &model,
                              int idx) {
  using namespace flexiblesusy::sm_twoloophiggs;
  const double p = 0.;
  const double mh2_shift_1l = calculate_Mh2_shift_1l_2l(model, idx);
  auto sm_2l = sm;
  const double self_energy_2l = Re(sm_2l.self_energy_hh_2loop(p));
  int nrows = sm.get_Yu().rows();
  int ncols = sm.get_Yu().cols();

  for (int i = 0; i < nrows; i++) {
    for (int k = 0; k < ncols; k++) {

      if (sm.HIGGS_2LOOP_CORRECTION_AT_AT && i == 2 && k == 2) {
        continue;
      }
      sm_2l.set_Yu(i, k, .0);
      sm_2l.set_Yd(i, k, .0);
    }
  }

  nrows = sm.get_Ye().rows();
  ncols = sm.get_Ye().cols();

  for (int i = 0; i < nrows; i++) {
    for (int k = 0; k < ncols; k++) {

      if (sm.HIGGS_2LOOP_CORRECTION_ATAU_ATAU && i == 2 && k == 2) {
        continue;
      }
      sm_2l.set_Ye(i, k, .0);
    }
  }
  sm_2l.calculate_DRbar_masses();

  //  tadpole are within selfenergy

  double mh2_shift =
      -self_energy_2l +
      (delta_mh_1loop_at_sm_deriv_p2(p, sm_2l.get_scale(), sm_2l.get_MFu(2),
                                     sm_2l.get_Yu(2, 2)) +
       delta_mh_1loop_ab_sm_deriv_p2(p, sm_2l.get_scale(), sm_2l.get_MFd(2),
                                     sm_2l.get_Yd(2, 2)) +
       delta_mh_1loop_atau_sm_deriv_p2(p, sm_2l.get_scale(), sm_2l.get_MFe(2),
                                       sm_2l.get_Ye(2, 2))) *
          (mh2_shift_1l);

  return mh2_shift;
}

double calculate_mh_DRbar_1l(const standard_model::Standard_model sm_0l,
                             const NUHMSSMNoFVHimalaya_mass_eigenstates &model,
                             int idx) {
  const double mh2_sm = Sqr(sm_0l.get_Mhh());
  const double Mh2_sm = calculate_Mh2_pole_1l(sm_0l);
  const double Mh2_bsm = calculate_Mh2_pole_1l(model, idx);

  const double mh2_1l = Mh2_bsm - Mh2_sm + mh2_sm;

  return AbsSqrt(mh2_1l);
}

/// calculates quartic Higgs coupling of SM at 0l
double calculate_lambda_0l(const NUHMSSMNoFVHimalaya_mass_eigenstates model,
                           const standard_model::Standard_model sm, int idx) {
  const double v = sm.get_v();
  const double lambda = Sqr(model.get_Mhh(idx)) / Sqr(v);

  return lambda;
}

/// calculates quartic Higgs coupling of SM at 1l precision
double calculate_lambda_1l(const NUHMSSMNoFVHimalaya_mass_eigenstates model,
                           const standard_model::Standard_model sm_0l,
                           standard_model::Standard_model sm_1l, int idx) {
  const double v = sm_0l.get_v();
  const double delta_v = (sm_1l.get_v()) - v;
  const double v2 = Sqr(sm_0l.get_v());
  const double mh2_SM = Sqr(calculate_mh_DRbar_1l(sm_0l, model, idx));

  const double lambda1l =
      mh2_SM / v2 - 2 * Sqr(model.get_Mhh(idx)) / v2 * delta_v / v;

  return lambda1l;
}

double calculate_MW_pole_1loop(const standard_model::Standard_model &sm_0l) {
  const double mw = sm_0l.get_MVWp();
  const double p = sm_0l.get_MVWp();
  const double self_energy = Re(sm_0l.self_energy_VWp_1loop(p));
  const double M_loop = Sqr(mw) - self_energy;

  return SignedAbsSqrt(M_loop);
}

double
calculate_MW_pole_1loop(const NUHMSSMNoFVHimalaya_mass_eigenstates &model_0l) {
  const double mw = model_0l.get_MVWm();
  const double p = model_0l.get_MVWm();
  const double self_energy = Re(model_0l.self_energy_VWm_1loop(p));
  const double M_loop = Sqr(mw) - self_energy;

  return SignedAbsSqrt(M_loop);
}

double calculate_MZ_pole_1loop(const standard_model::Standard_model &sm_0l) {
  const double mz = sm_0l.get_MVZ();
  const double p = sm_0l.get_MVZ();
  const double self_energy = Re(sm_0l.self_energy_VZ_1loop(p));
  const double M_loop = Sqr(mz) - self_energy;

  return SignedAbsSqrt(M_loop);
}

double
calculate_MZ_pole_1loop(const NUHMSSMNoFVHimalaya_mass_eigenstates &model_0l) {
  const double mz = model_0l.get_MVZ();
  const double p = model_0l.get_MVZ();
  const double self_energy = Re(model_0l.self_energy_VZ_1loop(p));
  const double M_loop = Sqr(mz) - self_energy;

  return SignedAbsSqrt(M_loop);
}

double
calculate_MW_MSbar_1loop(const standard_model::Standard_model &sm_0l,
                         const NUHMSSMNoFVHimalaya_mass_eigenstates &model_0l) {
  const double MW_sm = calculate_MW_pole_1loop(sm_0l);
  const double MW_bsm = calculate_MW_pole_1loop(model_0l);
  const double mw2 = Sqr(MW_bsm) - Sqr(MW_sm) + Sqr(sm_0l.get_MVWp());

  return AbsSqrt(mw2);
}

double
calculate_MZ_MSbar_1loop(const standard_model::Standard_model &sm_0l,
                         const NUHMSSMNoFVHimalaya_mass_eigenstates &model_0l) {

  const double MZ_sm = calculate_MZ_pole_1loop(sm_0l);
  const double MZ_bsm = calculate_MZ_pole_1loop(model_0l);
  const double mz2 = Sqr(MZ_bsm) - Sqr(MZ_sm) + Sqr(sm_0l.get_MVZ());

  return AbsSqrt(mz2);
}

/**
 * Calculates the gauge and Yukawa couplings and the VEV in the SM
 */
void match_high_to_low_scale_sm_tree_level(
    NUHMSSMNoFVHimalaya_mass_eigenstates &model_input,
    standard_model::Standard_model &sm, int idx) {
  auto model = model_input;

  sm.set_g1(model.get_g1());
  sm.set_g2(model.get_g2());
  sm.set_g3(model.get_g3());

  double v = Sqrt(Sqr(model.get_vu()) + Sqr(model.get_vd()));

  model.calculate_DRbar_masses();

  Eigen::Matrix<double, 3, 3> upQuarksDRbar = ZEROMATRIX(3, 3);
  Eigen::Matrix<double, 3, 3> downQuarksDRbar = ZEROMATRIX(3, 3);
  Eigen::Matrix<double, 3, 3> downLeptonsDRbar = ZEROMATRIX(3, 3);

  upQuarksDRbar = calculate_MFu_DRbar_tree_level(model);
  downQuarksDRbar = calculate_MFd_DRbar_tree_level(model);
  downLeptonsDRbar = calculate_MFe_DRbar_tree_level(model);

  sm.set_Yu((1.4142135623730951 * upQuarksDRbar) / v);
  sm.set_Yd((1.4142135623730951 * downQuarksDRbar) / v);
  sm.set_Ye((1.4142135623730951 * downLeptonsDRbar) / v);

  sm.set_v(v);

  sm.set_Lambdax(calculate_lambda_0l(model, sm, idx));
  sm.calculate_DRbar_masses();
}

void match_high_to_low_scale_sm_1l(NUHMSSMNoFVHimalaya_mass_eigenstates &model,
                                   standard_model::Standard_model &sm_0l,
                                   standard_model::Standard_model &sm_1l,
                                   int idx) {

  const double gy = sm_0l.get_g1() * standard_model_info::normalization_g1;
  const double g2 = sm_0l.get_g2() * standard_model_info::normalization_g2;

  const double gy_2 =
      Sqr(sm_0l.get_g1() * standard_model_info::normalization_g1);
  const double g2_2 =
      Sqr(sm_0l.get_g2() * standard_model_info::normalization_g2);
  const double g3_2 =
      Sqr(sm_0l.get_g3() * standard_model_info::normalization_g3);
  const double v = sm_0l.get_v();

  const double alpha_em = gy_2 * g2_2 / (4. * Pi * (gy_2 + g2_2));
  const double alpha_s = g3_2 / (4. * Pi);
  const double delta_alpha_em = calculate_delta_alpha_em(alpha_em, model);
  const double delta_alpha_s = calculate_delta_alpha_s(alpha_s, model);

  const double mW2_1l = Sqr(calculate_MW_MSbar_1loop(sm_0l, model));
  const double mZ2_1l = Sqr(calculate_MZ_MSbar_1loop(sm_0l, model));

  const double mW2 = Sqr(model.get_MVWm());
  const double mZ2 = Sqr(model.get_MVZ());

  const double delta_mZ2 = mZ2_1l - mZ2;
  const double delta_mW2 = mW2_1l - mW2;

  const double sThetaW2 = 1 - Sqr(sm_0l.get_MVWp() / sm_0l.get_MVZ());
  const double delta_sThetaW2 = delta_mZ2 * mW2 / (Sqr(mZ2)) - delta_mW2 / mZ2;

  const double g1_1l =
      AbsSqrt(4. * Pi * alpha_em * mZ2 / mW2) *
      (1. + 0.5 * (-delta_alpha_em + delta_sThetaW2 * mZ2 / mW2)) /
      standard_model_info::normalization_g1;
  const double g2_1l =
      AbsSqrt(4. * Pi * alpha_em / (sThetaW2)) *
      (1. + 0.5 * (-delta_alpha_em - delta_sThetaW2 / sThetaW2)) /
      standard_model_info::normalization_g2;
  const double g3_1l = AbsSqrt(4. * Pi * alpha_s) * (1. - 0.5 * delta_alpha_s) /
                       standard_model_info::normalization_g3;

  sm_1l.set_g1(g1_1l);
  sm_1l.set_g2(g2_1l);
  sm_1l.set_g3(g3_1l);

  const double delta_gy = g1_1l * standard_model_info::normalization_g1 - gy;
  const double delta_g2 = g2_1l * standard_model_info::normalization_g2 - g2;

  const double vev_1l =
      2 * Sqrt(mW2) / g2 * (1 - delta_g2 / g2 + 0.5 * delta_mW2 / mW2);
  const double delta_vev = vev_1l - v;

  Eigen::Matrix<double, 3, 3> YupQuarksMSbar = sm_0l.get_Yu();
  Eigen::Matrix<double, 3, 3> YdownQuarksMSbar = sm_0l.get_Yd();
  Eigen::Matrix<double, 3, 3> YdownLeptonsMSbar = sm_0l.get_Ye();

  Eigen::Matrix<double, 3, 3> upQuarksDRbar1l =
      calculate_MFu_DRbar_1loop(sm_0l, model);
  Eigen::Matrix<double, 3, 3> downQuarksDRbar1l =
      calculate_MFd_DRbar_1loop(sm_0l, model);
  Eigen::Matrix<double, 3, 3> downLeptonsDRbar1l =
      calculate_MFe_DRbar_1loop(sm_0l, model);

  sm_1l.set_Yu(1.4142135623730951 / v * upQuarksDRbar1l -
               YupQuarksMSbar * delta_vev / v);
  sm_1l.set_Yd(1.4142135623730951 / v * downQuarksDRbar1l -
               YdownQuarksMSbar * delta_vev / v);
  sm_1l.set_Ye(1.4142135623730951 / v * downLeptonsDRbar1l -
               YdownLeptonsMSbar * delta_vev / v);

  sm_1l.set_v(vev_1l);
  sm_1l.set_Lambdax(calculate_lambda_1l(model, sm_0l, sm_1l, idx));
  sm_1l.calculate_DRbar_masses();
}

// Mh ~ ATAU*ATAU,  AT*AS and AB*AS not independent
//
//  -> If AT*AS on , then consistency requires ATAU*ATAU
//
void match_high_to_low_scale_sm_2l(
    NUHMSSMNoFVHimalaya_mass_eigenstates &model,
    NUHMSSMNoFVHimalaya_mass_eigenstates &model_2l_gl,
    standard_model::Standard_model &sm_0l,
    standard_model::Standard_model &sm_0l_gl,
    standard_model::Standard_model &sm_1l,
    standard_model::Standard_model &sm_1l_gl,
    standard_model::Standard_model &sm_1l_gl_g3less,
    standard_model::Standard_model &sm_2l, int idx) {
  using namespace flexiblesusy::sm_twoloophiggs;

  const double lambda_1l = sm_1l.get_Lambdax();

  double delta_yt = 0.;
  double delta_yb = 0.;
  double delta_ytau = 0.;
  double delta_v_gl = 0.;
  double delta_mh_1l_gl = 0.;
  double mh2_conversion = 0.;

  // g3less refers to g3 = ytau = 0

  const double v = sm_0l_gl.get_v();
  const double v2 = Sqr(v);

  // threshold correction to v, yb, yt ~ O( at + ab )
  const double delta_v_gl_ytauless = sm_1l_gl_g3less.get_v() - v;
  const double delta_yt_g3ytauless =
      sm_1l_gl_g3less.get_Yu(2, 2) - sm_0l_gl.get_Yu(2, 2);
  const double delta_yb_g3ytauless =
      sm_1l_gl_g3less.get_Yd(2, 2) - sm_0l_gl.get_Yd(2, 2);
  const double p = 0.;

  if (sm_2l.HIGGS_2LOOP_CORRECTION_AT_AT) {
    delta_yt += delta_yt_g3ytauless;
    delta_yb += delta_yb_g3ytauless;
    delta_v_gl += delta_v_gl_ytauless;
  }

  if (sm_2l.HIGGS_2LOOP_CORRECTION_AT_AS) {
    delta_yt +=
        sm_1l_gl.get_Yu(2, 2) - sm_0l_gl.get_Yu(2, 2) - delta_yt_g3ytauless;
  }

  if (sm_2l.HIGGS_2LOOP_CORRECTION_AB_AS) {
    delta_yb +=
        sm_1l_gl.get_Yd(2, 2) - sm_0l_gl.get_Yd(2, 2) - delta_yb_g3ytauless;
  }

  if (sm_2l.HIGGS_2LOOP_CORRECTION_ATAU_ATAU) {
    delta_ytau += sm_1l_gl.get_Ye(2, 2) - sm_0l_gl.get_Ye(2, 2);
    delta_v_gl += sm_1l_gl.get_v() - v - delta_v_gl_ytauless;
  }

  model.calculate_DRbar_masses();
  const double mh2_bsm_shift = calculate_Mh2_shift_2l(model_2l_gl, idx);
  const double mh2_sm_shift =
      calculate_Mh2_shift_2l(sm_0l_gl, model_2l_gl, idx);
  const double mh2_1l_sm_shift = calculate_Mh2_shift_1l_2l(sm_0l_gl);
  const double mh2_1l_bsm_shift = calculate_Mh2_shift_1l_2l(model_2l_gl, idx);

  mh2_conversion +=
      delta_mh_1loop_at_sm_deriv_yt(
          p, sm_0l_gl.get_scale(), sm_0l_gl.get_MFu(2), sm_0l_gl.get_Yu(2, 2)) *
          delta_yt +
      delta_mh_1loop_ab_sm_deriv_yb(
          p, sm_0l_gl.get_scale(), sm_0l_gl.get_MFd(2), sm_0l_gl.get_Yd(2, 2)) *
          delta_yb +
      delta_mh_1loop_atau_sm_deriv_ytau(
          p, sm_0l_gl.get_scale(), sm_0l_gl.get_MFe(2), sm_0l_gl.get_Ye(2, 2)) *
          delta_ytau +
      (delta_mh_1loop_at_sm_deriv_v(p, sm_0l_gl.get_scale(),
                                    sm_0l_gl.get_MFu(2),
                                    sm_0l_gl.get_Yu(2, 2)) +
       delta_mh_1loop_ab_sm_deriv_v(p, sm_0l_gl.get_scale(),
                                    sm_0l_gl.get_MFd(2),
                                    sm_0l_gl.get_Yd(2, 2)) +
       delta_mh_1loop_atau_sm_deriv_v(p, sm_0l_gl.get_scale(),
                                      sm_0l_gl.get_MFe(2),
                                      sm_0l_gl.get_Ye(2, 2))) *
          delta_v_gl;

  const double lambda_2l =
      lambda_1l + (mh2_bsm_shift - mh2_sm_shift - mh2_conversion) / v2 -
      2. * delta_v_gl / (v * v2) * (mh2_1l_bsm_shift - mh2_1l_sm_shift);

  const double delta_vev = sm_1l.get_v() - v;
  const double yt_0l = sm_0l.get_Yu(2, 2);

  const double mt_2l = calculate_MFt_MSbar_2loop(sm_0l, model);

  sm_2l.set_Yu(2, 2, (Sqrt(2) * mt_2l / v - yt_0l * delta_vev / v));
  sm_2l.set_Lambdax(lambda_2l);
  sm_2l.calculate_DRbar_masses();
}

// Consistent cancelation of large logs only for
//
//     loop_corrections.top_qcd  > 1
//
// has to be ensured!

void match_high_to_low_scale_sm_3l(
    NUHMSSMNoFVHimalaya_mass_eigenstates &model_gl,
    standard_model::Standard_model &sm_0l_gl,
    standard_model::Standard_model &sm_1l,
    standard_model::Standard_model &sm_1l_gl,
    standard_model::Standard_model &sm_1l_gl_g3less,
    standard_model::Standard_model &sm_2l,
    standard_model::Standard_model &sm_3l, int idx) {
  using namespace flexiblesusy::sm_twoloophiggs;

  const double lambda_2l = sm_2l.get_Lambdax();
  double delta_lambda_3l = 0.;

  double v2 = Sqr(sm_0l_gl.get_v());
  double yt = sm_0l_gl.get_Yu(2, 2);
  double yt2 = Sqr(yt);
  double gs = sm_0l_gl.get_g3();
  double gs2 = Sqr(gs);
  double gs4 = Sqr(gs2);
  double MS = Sqrt(model_gl.get_MSt(0) * model_gl.get_MSt(1));
  double Q2 = Sqr(sm_0l_gl.get_scale());
  double mt = model_gl.get_MFt();
  double mt2 = Sqr(mt);
  double logmt = Log(mt2 / Q2);
  double logmt2 = Sqr(Log(mt2 / Q2));
  double logmt3 = logmt * logmt2;
  double k = 1. / Power(4 * Pi, 2);
  double k2 = 1. / Power(4 * Pi, 4);
  double k3 = 1. / Power(4 * Pi, 6);

  double delta_yt_1l = sm_1l_gl.get_Yu(2, 2) - sm_1l_gl_g3less.get_Yu(2, 2);
  double delta_yt_2l = sm_2l.get_Yu(2, 2) - sm_1l.get_Yu(2, 2);
  double sqr_delta_yt_1l = Sqr(delta_yt_1l);
  double delta_g3_1l = sm_1l.get_g3() - sm_0l_gl.get_g3();

  double S1_deriv_yt = delta_mh_1loop_at_sm_deriv_yt(0, sm_0l_gl.get_scale(),
                                                     sm_0l_gl.get_MFu(2),
                                                     sm_0l_gl.get_Yu(2, 2)) *
                       delta_yt_2l;
  double S1_deriv_yt2 =
      -0.5 * (24 * k * mt2 * (7. + 6. * logmt)) * sqr_delta_yt_1l;
  double S2_deriv_yt =
      (64. * k2 * gs2 * mt2 * yt * (1. + 8. * logmt + 6. * logmt2)) *
      delta_yt_1l;
  double S2_deriv_gs =
      (64. * k2 * gs * mt2 * yt2 * logmt * (1. + 3. * logmt)) * delta_g3_1l;

  double mh2_conversion =
      S1_deriv_yt + S1_deriv_yt2 + S2_deriv_yt + S2_deriv_gs;
  double mh2_sm_shift = -Re(sm_0l_gl.self_energy_hh_3loop());

  // calculate 2L self-energy using tree-level parameters
  Eigen::Matrix<double, 2, 2> self_energy_3l(
      Eigen::Matrix<double, 2, 2>::Zero());

  try {
    self_energy_3l = Re(model_gl.self_energy_hh_3loop());

    const auto Mh2_loop =
        (calculate_mh2_tree_level(model_gl) - self_energy_3l).eval();
    Eigen::Array<double, 2, 1> Mh2_pole;
    fs_diagonalize_hermitian(Mh2_loop, Mh2_pole);
    double mh2_bsm_shift = Mh2_pole(idx) - Sqr(model_gl.get_Mhh(idx));

    delta_lambda_3l = (mh2_bsm_shift - mh2_sm_shift - mh2_conversion) / v2;

  } catch (const flexiblesusy::Error &e) {
    //  WARNING("2-loop Higgs mass calculation failed: " << e.what());
  }

  const double lambda_3l = lambda_2l + delta_lambda_3l;
  sm_3l.set_Lambdax(lambda_3l);
  sm_3l.calculate_DRbar_masses();
}

/**
 * Calculates SM tree-level parameters (g1, g2, g3, Yu, Yd,
 * Ye, v) by performing a tree-level matching from the MSSM to the
 * SM
 *
 * @param model MSSM parameters
 * @param sm SM parameters
 *
 * @return SM with tree-level parameters
 */
standard_model::Standard_model
calculate_SM_tree_level(const NUHMSSMNoFVHimalaya_mass_eigenstates &model,
                        int index) {
  standard_model::Standard_model sm_0l;
  sm_0l.set_precision(model.get_precision());
  sm_0l.set_scale(model.get_scale());

  auto model_0l = model;
  model_0l.get_problems().clear();
  model_0l.calculate_DRbar_masses();

  match_high_to_low_scale_sm_tree_level(model_0l, sm_0l, index);

  return sm_0l;
}

standard_model::Standard_model
calculate_SM_1l(const NUHMSSMNoFVHimalaya_mass_eigenstates &model,
                const standard_model::Standard_model &sm, int index) {
  standard_model::Standard_model sm_1l;
  sm_1l.set_precision(model.get_precision());
  sm_1l.set_scale(model.get_scale());

  auto sm_0l = sm;
  sm_0l.get_problems().clear();
  sm_0l.calculate_DRbar_masses();

  auto model_1l = model;
  model_1l.get_problems().clear();
  model_1l.calculate_DRbar_masses();

  match_high_to_low_scale_sm_1l(model_1l, sm_0l, sm_1l, index);

  return sm_1l;
}

standard_model::Standard_model
calculate_SM_1l(const NUHMSSMNoFVHimalaya_mass_eigenstates &model,
                const standard_model::Standard_model &sm, double gauge_less,
                bool g3_ytau_less, int index) {

  standard_model::Standard_model sm_0l;
  sm_0l.set_precision(model.get_precision());
  sm_0l.set_scale(model.get_scale());

  auto model_0l = model;
  model_0l.get_problems().clear();
  model_0l.set_g1(gauge_less / NUHMSSMNoFVHimalaya_info::normalization_g1);
  model_0l.set_g2(gauge_less / NUHMSSMNoFVHimalaya_info::normalization_g2);
  if (g3_ytau_less == True) {
    Eigen::Matrix<double, 3, 3> Zeromat;
    Zeromat.setZero();
    model_0l.set_Ye(Zeromat);
    model_0l.set_TYe(Zeromat);
    model_0l.set_g3(gauge_less);
  }
  model_0l.calculate_DRbar_masses();

  match_high_to_low_scale_sm_tree_level(model_0l, sm_0l, index);

  standard_model::Standard_model sm_1l;
  sm_1l.set_precision(model.get_precision());
  sm_1l.set_scale(model.get_scale());

  match_high_to_low_scale_sm_1l(model_0l, sm_0l, sm_1l, index);

  return sm_1l;
}

standard_model::Standard_model
calculate_SM_2l(const NUHMSSMNoFVHimalaya_mass_eigenstates &model,
                const standard_model::Standard_model &sm_0L,
                const standard_model::Standard_model &sm_1L,
                const standard_model::Standard_model &sm_1L_GL,
                const standard_model::Standard_model &sm_1L_GL_G3_Ytau_Less,
                double gauge_less, int index) {
  auto sm_2l = sm_1L;
  sm_2l.set_precision(model.get_precision());
  sm_2l.set_loop_corrections(model.get_loop_corrections());


  auto sm_0l = sm_0L;
  auto sm_1l = sm_1L;
  auto sm_1l_gl = sm_1L_GL;
  auto sm_1l_gl_g3_ytau_less = sm_1L_GL_G3_Ytau_Less;
  auto model_normal = model;
  model_normal.calculate_DRbar_masses();

  standard_model::Standard_model sm_0l_gl;
  sm_0l_gl.set_precision(model.get_precision());
  sm_0l_gl.set_scale(model.get_scale());

  auto model_2l_gl = model;
  model_2l_gl.get_problems().clear();
  model_2l_gl.set_g1(gauge_less / NUHMSSMNoFVHimalaya_info::normalization_g1);
  model_2l_gl.set_g2(gauge_less / NUHMSSMNoFVHimalaya_info::normalization_g2);
  model_2l_gl.calculate_DRbar_masses();

  match_high_to_low_scale_sm_tree_level(model_2l_gl, sm_0l_gl, index);

  sm_0l_gl.set_loop_corrections(model.get_loop_corrections());
  match_high_to_low_scale_sm_2l(model_normal, model_2l_gl, sm_0l, sm_0l_gl,
                                sm_1l, sm_1l_gl, sm_1l_gl_g3_ytau_less, sm_2l,
                                index);

  return sm_2l;
}

standard_model::Standard_model
calculate_SM_3l(const NUHMSSMNoFVHimalaya_mass_eigenstates &model,
                const standard_model::Standard_model &sm_1L,
                const standard_model::Standard_model &sm_1L_GL,
                const standard_model::Standard_model &sm_2L, double gauge_less,
                int index) {
  auto sm_3l = sm_2L;
  sm_3l.set_precision(model.get_precision());
  sm_3l.set_loop_corrections(model.get_loop_corrections());

  auto sm_1l = sm_1L;
  auto sm_2l = sm_2L;

  auto model_normal = model;
  model_normal.calculate_DRbar_masses();

  standard_model::Standard_model sm_0l_gl;
  sm_0l_gl.set_precision(model.get_precision());
  sm_0l_gl.set_scale(model.get_scale());

  auto model_gl = model;
  model_gl.get_problems().clear();
  model_gl.set_g1(gauge_less / NUHMSSMNoFVHimalaya_info::normalization_g1);
  model_gl.set_g2(gauge_less / NUHMSSMNoFVHimalaya_info::normalization_g2);
  model_gl.set_Yd(2, 2, 0.);
  model_gl.set_Ye(2, 2, 0.);
  Eigen::Matrix<double, 3, 3> TYd, TYe;
  TYd.setZero();
  TYe.setZero();
  model_gl.set_TYd(TYd);
  model_gl.set_TYe(TYe);
  model_gl.calculate_DRbar_masses();

  match_high_to_low_scale_sm_tree_level(model_gl, sm_0l_gl, index);
  auto sm_1l_gl_g3_ytau_less =
      calculate_SM_1l(model_gl, sm_0l_gl, gauge_less, 1, index);
  auto sm_1l_gl = calculate_SM_1l(model_gl, sm_0l_gl, gauge_less, 0, index);

  sm_0l_gl.set_loop_corrections(model.get_loop_corrections());
  match_high_to_low_scale_sm_3l(model_gl, sm_0l_gl, sm_1l, sm_1l_gl,
                                sm_1l_gl_g3_ytau_less, sm_2l, sm_3l, index);

  return sm_3l;
}

// calculates lambda and runs down to Q_low
standard_model::Standard_model
run_down(const Spectrum_generator_settings &settings,
         NUHMSSMNoFVHimalaya_mass_eigenstates &model,
         const NUHMSSMNoFVHimalaya_input_parameters &mssm_input, double Q_low,
         double gauge_less) {
  const int matching_loop_order_down =
      settings.get(Spectrum_generator_settings::eft_matching_loop_order_down);
  const int matching_loop_order_yuk =
      settings.get(Spectrum_generator_settings::eft_matching_loop_order_up);
  const int index = settings.get(Spectrum_generator_settings::eft_higgs_index);
  const int DeltaLam = mssm_input.DeltaOS;

  auto sm_0l = calculate_SM_tree_level(model, index);

  auto sm_1l = calculate_SM_1l(model, sm_0l, index);
  auto sm_1l_gl_g3_ytau_less =
      calculate_SM_1l(model, sm_0l, gauge_less, 1, index);
  auto sm_1l_gl = calculate_SM_1l(model, sm_0l, gauge_less, 0, index);
  auto sm_2l = calculate_SM_2l(model, sm_0l, sm_1l, sm_1l_gl,
                               sm_1l_gl_g3_ytau_less, gauge_less, index);

  switch (matching_loop_order_down) {
  case 0: {
    fill_settings(sm_0l, settings);
    sm_0l.run_to(Q_low);
    return sm_0l;
  }
  case 1: {
    fill_settings(sm_1l, settings);

    if (matching_loop_order_yuk == 0) {
      sm_1l.set_Yu(sm_0l.get_Yu());
      sm_1l.set_Yd(sm_0l.get_Yd());
      sm_1l.set_Ye(sm_0l.get_Ye());
    }
    if (matching_loop_order_yuk == 2) {
      sm_1l.set_Yu(sm_2l.get_Yu());
    }
    sm_1l.calculate_DRbar_masses();
    sm_1l.run_to(Q_low);
    return sm_1l;
  }
  case 2: {
    fill_settings(sm_2l, settings);
    if (matching_loop_order_yuk == 0) {
      sm_2l.set_Yu(sm_0l.get_Yu());
      sm_2l.set_Yd(sm_0l.get_Yd());
      sm_2l.set_Ye(sm_0l.get_Ye());
    }
    if (matching_loop_order_yuk == 1) {
      sm_2l.set_Yu(sm_1l.get_Yu());
      sm_2l.set_Yd(sm_1l.get_Yd());
      sm_2l.set_Ye(sm_1l.get_Ye());
    }

    if (DeltaLam == 1) {
      sm_2l.set_Lambdax(sm_2l.get_Lambdax() +
                        calculate_delta_lam_2l(model, sm_0l, sm_1l, sm_1l_gl));
    }
    sm_2l.calculate_DRbar_masses();
    sm_2l.set_integrator(runge_kutta::RKF_integrator());
    sm_2l.run_to(Q_low);
    return sm_2l;
  }
  case 3: {
    auto sm_3l =
        calculate_SM_3l(model, sm_1l, sm_1l_gl, sm_2l, gauge_less, index);

    fill_settings(sm_3l, settings);
    if (matching_loop_order_yuk == 0) {
      sm_3l.set_Yu(sm_0l.get_Yu());
      sm_3l.set_Yd(sm_0l.get_Yd());
      sm_3l.set_Ye(sm_0l.get_Ye());
    }
    if (matching_loop_order_yuk == 1) {
      sm_3l.set_Yu(sm_1l.get_Yu());
      sm_3l.set_Yd(sm_1l.get_Yd());
      sm_3l.set_Ye(sm_1l.get_Ye());
    }
    if (DeltaLam == 1) {
      sm_3l.set_Lambdax(sm_3l.get_Lambdax() +
                        calculate_delta_lam_3l(model, sm_0l, sm_1l, sm_1l_gl));
    }

    sm_3l.calculate_DRbar_masses();
    sm_3l.set_integrator(runge_kutta::RKF_integrator());
    sm_3l.run_to(Q_low);
    return sm_3l;
  }
  default:
    std::cout << "invalid loop order down" << std::endl;
  }

  return sm_0l;
}

double calc_SUSY_scale(NUHMSSMNoFVHimalaya_mass_eigenstates &model) {
  model.calculate_DRbar_masses();

  const auto MSt = model.get_MSt();

  return AbsSqrt(MSt(0) * MSt(1));
}

/// resets MSSM parameters and runs down
standard_model::Standard_model
run_down_with(const Spectrum_generator_settings &settings,
              const NUHMSSMNoFVHimalaya_input_parameters &mssm_input,
              const Vec_t &mssm_pars, double Q_match, double Q_low,
              double gauge_less) {
  // create MSSM object
  NUHMSSMNoFVHimalaya_mass_eigenstates mssm;
  fill_parameters(mssm, mssm_pars);  // fill SM-like parameters
  fill_parameters(mssm, mssm_input); // fill SUSY input parameters
  fill_settings(mssm, settings);

  obtain_parameters(mssm, high_scale_pars);

  mssm.calculate_DRbar_masses();
  mssm.solve_ewsb();

  mssm.set_integrator(runge_kutta::RKF_integrator());
  // run to matching scale
  mssm.run_to(Q_match);
  mssm.solve_ewsb();

  // calculate lambda and run down
  return run_down(settings, mssm, mssm_input, Q_low, gauge_less);
}

softsusy::QedQcd
predicted_pole_masses(const standard_model::Standard_model &sm) {
  softsusy::QedQcd qedqcd;
  qedqcd.setPoleMt(sm.get_physical().MFu(2));
  qedqcd.setPoleMW(sm.get_physical().MVWp);
  qedqcd.setPoleMZ(sm.get_physical().MVZ);

  return qedqcd;
}

/// calculates the deviation between observables calculated from the
/// SM parameters and given input values (`sm_input')
Vec_t deviation_from_observables(
    const Spectrum_generator_settings &spectrum_generator_settings,
    const NUHMSSMNoFVHimalaya_input_parameters &mssm_input,
    const softsusy::QedQcd &sm_input, const Vec_t &mssm_pars, double Q_match,
    double gauge_less) {

  auto sm = run_down_with(spectrum_generator_settings, mssm_input, mssm_pars,
                          Q_match, sm_input.displayPoleMZ(), gauge_less);
  // calculate observables

  /*

    1) HSSUSY treatment of Mtop at Qlow:
    comment in this block
    sm.calculate_pole_masses_rotated(Mtop) line 2) comment out lines commented
    as "automatized SG treatment of Mtop "

     auto sm_QEDQCD           = sm_input;
     sm_QEDQCD.toMz();
     double Mtop           = sm_input.displayPoleMt();
  */
  sm.solve_ewsb();
  sm.calculate_pole_masses(); // automatized SG treatment of Mtop

  // sm.calculate_pole_masses_rotated(Mtop); //HSSUSY treatment of Mtop

  Vec_t diff(Vec_t::Zero());
  int idx = 0;

  const double gY = sm.get_g1() * standard_model_info::normalization_g1;
  const double g2 = sm.get_g2() * standard_model_info::normalization_g2;
  const double g3 = sm.get_g3() * standard_model_info::normalization_g3;
  const double e = gY * g2 / Sqrt(Sqr(gY) + Sqr(g2));
  const double alpha_em_msbar = Sqr(e) / (4 * Pi);
  const double alpha_s_msbar = Sqr(g3) / (4 * Pi);
  const auto qedqcd = predicted_pole_masses(sm);
  const double cos_theta = sm.get_MVWp() / sm.get_MVZ();
  const double g_fermi = sm.calculate_G_fermi(qedqcd);

  auto sm_QEDQCD = sm_input; // automatized SG treatment of Mtop
  sm_QEDQCD.toMz();          // automatized SG treatment of Mtop
  const double mb_QEDQCD = sm_QEDQCD.displayMass(softsusy::mBottom);
  const double mc_QEDQCD = sm_QEDQCD.displayMass(softsusy::mCharm);
  const double ms_QEDQCD = sm_QEDQCD.displayMass(softsusy::mStrange);

  const double alpha_em_SM5 = sm_input.displayAlpha(softsusy::ALPHA);
  const double alpha_s_SM5 = sm_input.displayAlpha(softsusy::ALPHAS);

  const double delta_alpha_em = sm.calculate_delta_alpha_em(alpha_em_msbar);
  const double delta_alpha_s = sm.calculate_delta_alpha_s(alpha_s_SM5);
  const double Mtop =
      sm_input.displayPoleMt(); // automatized SG treatment of Mtop
  const double yt_MZ = Electroweak_constants::ytSM;
  //		              sm_input.displayAlpha(softsusy::ALPHAS));

  const double alpha_em_SM5_p = alpha_em_msbar / (1 + delta_alpha_em);
  const double alpha_s_SM5_p = alpha_s_msbar * (1 - delta_alpha_s);

  diff(idx++) = (alpha_em_SM5 - alpha_em_SM5_p) / alpha_em_SM5;
  diff(idx++) = (alpha_s_SM5 - alpha_s_SM5_p) / alpha_s_SM5;
  diff(idx++) = (sm_input.displayFermiConstant() - g_fermi) /
                sm_input.displayFermiConstant();
  diff(idx++) = (Mtop - sm.get_physical().MFu(2)) / Mtop;
  diff(idx++) = (sm_input.displayPoleMZ() - sm.get_physical().MVZ) /
                sm_input.displayPoleMZ();
  diff(idx++) = (sm.calculate_MFd_DRbar(mb_QEDQCD, 2) - sm.get_MFd(2)) /
                sm.calculate_MFd_DRbar(mb_QEDQCD, 2);
  diff(idx++) = (sm_input.displayPoleMtau() - sm.get_physical().MFe(2)) /
                (sm_input.displayPoleMtau());

  /*
     std::cout << "----observables----"<< std::endl;
     std::cout << "alpha_em = " <<   alpha_em_SM5<< ",  Delta = "<<
     (alpha_em_SM5 - alpha_em_SM5_p)/alpha_em_SM5<<std::endl; std::cout <<
     "alpha_s  = " <<   alpha_s_SM5<< ",  Delta = "<<  (alpha_s_SM5 -
     alpha_s_SM5_p)/alpha_s_SM5<<std::endl;

     std::cout << "Gf       = " << g_fermi << ",  Delta = "<<
     (sm_input.displayFermiConstant() -
     g_fermi)/sm_input.displayFermiConstant()<<std::endl; std::cout << "Mz = "
     << sm.get_physical().MVZ<< ",  Delta = "<<  (sm_input.displayPoleMZ() -
     sm.get_physical().MVZ)/sm_input.displayPoleMZ()<<std::endl; std::cout <<
     "Mt       = " << sm.get_physical().MFu(2)<< ",  Delta = "<<
     (sm_input.displayPoleMt()-sm.get_physical().MFu(2))/Mtop<<std::endl;
     std::cout << "Mb       = " << sm.get_physical().MFd(2)<< ",  Delta = "<<
     (sm.calculate_MFd_DRbar(mb_QEDQCD, 2) -
     sm.get_MFd(2))/sm.get_MFd(2)<<std::endl; std::cout << "Mta      = " <<
     sm.get_physical().MFe(2)<< ",  Delta = "<<
     (sm_input.displayPoleMtau()-sm.get_physical().MFe(2))/(sm_input.displayPoleMtau())<<std::endl;
     std::cout<< "Iteration = "<< Iteration << std::endl;
  */
  double abs_diff = 0.;
  for (idx = 0; idx < NPARS; ++idx) {
    abs_diff += Abs(diff(idx++));
  }

  abs_deviation = abs_diff;

  return diff;
}

/// FlexibleEFTHiggs iteration between high and low scale
standard_model::Standard_model
run_FlexibleEFTHiggs(const Spectrum_generator_settings &settings,
                     const NUHMSSMNoFVHimalaya_input_parameters &mssm_input,
                     const softsusy::QedQcd &sm_input, double gauge_less) {
  const double precision_goal = 1e-5; // 1./3.;//1e-2;
  const int max_it = -5 * std::log10(precision_goal);

  // calculates deviation from input SM observables
  // -> need to find the root of this function

  Vec_t mssm_pars = to_vec(initial_guess(mssm_input));

  NUHMSSMNoFVHimalaya_mass_eigenstates mssm;
  fill_parameters(mssm, mssm_pars);  // fill SM-like parameters
  fill_parameters(mssm, mssm_input); // fill SUSY input parameters
  fill_settings(mssm, settings);

  const double susy_scale = calc_SUSY_scale(mssm);
  const double Q_match =
      settings.get(Spectrum_generator_settings::eft_matching_scale) != 0.
          ? settings.get(Spectrum_generator_settings::eft_matching_scale)
          : susy_scale;

  double Q_low =
      settings.get(Spectrum_generator_settings::eft_pole_mass_scale) != 0.
          ? settings.get(Spectrum_generator_settings::eft_pole_mass_scale)
          : sm_input.displayPoleMt();

  // std::cout<<"Q_match = "<< Q_match<<std::endl;
  auto dev = [&settings, &mssm_input, &sm_input, Q_match,
              gauge_less](const Vec_t &vec) { // TODO: what happens
    return deviation_from_observables(settings, mssm_input, sm_input, vec,
                                      Q_match, gauge_less);
  };

  Root_finder<NPARS> root_finder(dev, max_it, precision_goal,
                                 Root_finder<NPARS>::GSLHybridS);
  auto err = root_finder.solve(to_vec(initial_guess(mssm_input)));

  Vec_t pars_solution = root_finder.get_solution();

  if (err != EWSB_solver::SUCCESS) {
    Vec_t high_scale_pars_HybridS = high_scale_pars;
    double deviation_HybridS = abs_deviation;

    Root_finder<NPARS> root_finder_Hybrid(dev, max_it, precision_goal,
                                          Root_finder<NPARS>::GSLHybrid);
    err = root_finder_Hybrid.solve(high_scale_pars);
    pars_solution = root_finder_Hybrid.get_solution();

    if (err != EWSB_solver::SUCCESS) {
      Vec_t high_scale_pars_Hybrid = high_scale_pars;
      double deviation_Hybrid = abs_deviation;
      if (deviation_HybridS < deviation_Hybrid) {
        high_scale_pars = high_scale_pars_Hybrid;
      }

      Root_finder<NPARS> root_finder_Broyden(dev, max_it, precision_goal,
                                             Root_finder<NPARS>::GSLBroyden);
      err = root_finder_Broyden.solve(high_scale_pars);
      pars_solution = root_finder_Broyden.get_solution();
      if (err != EWSB_solver::SUCCESS) {

        throw std::runtime_error("-");
      }
    }
  }

  // run down for the last time with the solution that has been found
  auto sm = run_down_with(settings, mssm_input, pars_solution, Q_match, Q_low,
                          gauge_less);

  sm.solve_ewsb();
  sm.calculate_Mhh_pole();

  return sm;
}

NUHMSSMNoFVHimalaya_input_parameters make_mssm_input(double MS, double Xt,
                                                     double tb) {
  const double mu = MS;

  NUHMSSMNoFVHimalaya_input_parameters pars;
  pars.TanBeta = tb;
  pars.Qin = MS;
  pars.M1 = MS;
  pars.M2 = MS;
  pars.M3 = MS;
  pars.AtIN = Xt + mu / tb;
  pars.AbIN = 0;   // mu*tb;
  pars.AtauIN = 0; // mu*tb;
  pars.AcIN = 0;
  pars.AsIN = 0;
  pars.AmuonIN = 0;
  pars.AuIN = 0;
  pars.AdIN = 0;
  pars.AeIN = 0;
  pars.MuIN = mu;
  pars.mA2IN = Sqr(MS);
  pars.ml11IN = MS;
  pars.ml22IN = MS;
  pars.ml33IN = MS;
  pars.me11IN = MS;
  pars.me22IN = MS;
  pars.me33IN = MS;
  pars.mq11IN = MS;
  pars.mq22IN = MS;
  pars.mq33IN = MS;
  pars.mu11IN = MS;
  pars.mu22IN = MS;
  pars.mu33IN = MS;
  pars.md11IN = MS;
  pars.md22IN = MS;
  pars.md33IN = MS;

  return pars;
}

} // namespace flexiblesusy

int main(int argc, char *argv[]) {
  using namespace flexiblesusy;

  Command_line_options options(argc, argv);
  if (options.must_print_model_info())
    NUHMSSMNoFVHimalaya_info::print(std::cout);
  if (options.must_exit())
    return options.status();

  const std::string rgflow_file(options.get_rgflow_file());
  const std::string slha_input_source(options.get_slha_input_file());
  NUHMSSMNoFVHimalaya_slha_io slha_io;
  Spectrum_generator_settings spectrum_generator_settings;

  if (slha_input_source.empty()) {
    ERROR("No SLHA input source given!\n"
          "   Please provide one via the option --slha-input-file=");
    return EXIT_FAILURE;
  }

  Physical_input physical_input;
  NUHMSSMNoFVHimalaya_input_parameters input;

  // connecting to SLHA-input file

  try {
    slha_io.read_from_source(slha_input_source);
    slha_io.fill(input);
    slha_io.fill(physical_input);
    slha_io.fill(spectrum_generator_settings);
  } catch (const Error &error) {
    ERROR(error.what());
    return EXIT_FAILURE;
  }

  const double gauge_less = GAUGELESS;
  const auto mssm_input = input;
  const auto sm_input = make_sm_input();

  try {
    const auto sm = run_FlexibleEFTHiggs(spectrum_generator_settings,
                                         mssm_input, sm_input, gauge_less);

    std::cout << std::setprecision(9) << sm.get_physical().Mhh << '\n';
  } catch (const std::exception &e) {
    std::cerr << e.what() << '\n';
  }

  return 0;
}
